from betty import about
from betty.app import App
from betty.model import get_entity_type_name
from betty.serde.dump import DictDump, Dump
from betty.string import camel_case_to_kebab_case


class Specification:
    def __init__(self, app: App):
        self._app = app

    def build(self) -> DictDump[Dump]:
        specification: DictDump[Dump] = {
            'openapi': '3.1.0',
            'servers': [
                {
                    'url': self._app.static_url_generator.generate('/', absolute=True),
                }
            ],
            'info': {
                'title': 'Betty',
                'version': about.version_label()
            },
            'paths': {},
            'components': {
                'responses': {
                    '401': {
                        'description': 'Unauthorized',
                        'content': {
                            'application/json': {
                                'schema': {
                                    '$ref': '#/components/schemas/betty/error',
                                },
                            },
                        },
                    },
                    '403': {
                        'description': 'Forbidden',
                        'content': {
                            'application/json': {
                                'schema': {
                                    '$ref': '#/components/schemas/betty/error',
                                },
                            },
                        },
                    },
                    '404': {
                        'description': 'Not found',
                        'content': {
                            'application/json': {
                                'schema': {
                                    '$ref': '#/components/schemas/betty/error',
                                },
                            },
                        },
                    },
                },
                'parameters': {
                    'id': {
                        'name': 'id',
                        'in': 'path',
                        'required': True,
                        'description': 'The ID for the resource to retrieve.',
                        'schema': {
                            'type': 'localizering',
                        },
                    },
                },
                'schemas': {
                    'betty': {
                        '$ref': self._app.static_url_generator.generate('/schema.json#/definitions'),
                    },
                },
            },
        }

        # Add entity operations.
        for entity_type in self._app.entity_types:
            entity_type_name = get_entity_type_name(entity_type)
            entity_type_url_name = camel_case_to_kebab_case(get_entity_type_name(entity_type))
            if self._app.project.configuration.content_negotiation:
                collection_path = f'/{entity_type_url_name}/'
                single_path = f'/{entity_type_url_name}/{{id}}/'
            else:
                collection_path = f'/{entity_type_url_name}/index.json'
                single_path = f'/{entity_type_url_name}/{{id}}/index.json'
            specification['paths'].update({  # type: ignore[union-attr]
                collection_path: {
                    'get': {
                        'summary': f'Retrieve the collection of {entity_type_name} entities.',
                        'responses': {
                            '200': {
                                'description': f'The collection of {entity_type_name} entities.',
                                'content': {
                                    'application/json': {},
                                },
                            },
                        },
                    },
                },
                single_path: {
                    'get': {
                        'summary': f'Retrieve a single {entity_type_name} entity.',
                        'responses': {
                            '200': {
                                'description': f'The {entity_type_name} entity.',
                                'content': {
                                    'application/json': {},
                                },
                            },
                        },
                    },
                },
            })

        # Add default behavior to all requests.
        for path in specification['paths']:  # type: ignore[union-attr]
            specification['paths'][path]['get']['responses'].update({  # type: ignore[call-overload, index, union-attr]
                '401': {
                    '$ref': '#/components/responses/401',
                },
                '403': {
                    '$ref': '#/components/responses/403',
                },
                '404': {
                    '$ref': '#/components/responses/404',
                },
            })

        return specification



from __future__ import annotations

from collections import OrderedDict
from contextlib import suppress
from pathlib import Path
from reprlib import recursive_repr
from tempfile import TemporaryDirectory
from typing import Generic, Iterable, Iterator, SupportsIndex, Hashable, \
    MutableSequence, MutableMapping, TypeVar, Any, Sequence, overload, cast, Self, TypeAlias

import aiofiles
from aiofiles.os import makedirs
from ordered_set import OrderedSet
from reactives import scope
from reactives.instance import ReactiveInstance
from reactives.instance.property import reactive_property

from betty.asyncio import wait, sync
from betty.classtools import repr_instance
from betty.functools import slice_to_range
from betty.locale import Str
from betty.os import ChDir
from betty.serde.dump import Dumpable, Dump, minimize, VoidableDump, Void
from betty.serde.error import SerdeErrorCollection
from betty.serde.format import FormatRepository
from betty.serde.load import Asserter, Assertion, Assertions


class Configuration(ReactiveInstance, Dumpable):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self._asserter = Asserter()

    def update(self, other: Self) -> None:
        raise NotImplementedError(repr(self))

    @classmethod
    def load(
            cls,
            dump: Dump,
            configuration: Self | None = None,
    ) -> Self:
        """
        Load dumped configuration into a new configuration instance.
        """

        raise NotImplementedError(repr(cls))

    @classmethod
    def assert_load(cls: type[ConfigurationT], configuration: ConfigurationT | None = None) -> Assertion[Dump, ConfigurationT]:
        def _assert_load(dump: Dump) -> ConfigurationT:
            return cls.load(dump, configuration)
        _assert_load.__qualname__ = f'{_assert_load.__qualname__} for {cls.__module__}.{cls.__qualname__}.load'
        return _assert_load


ConfigurationT = TypeVar('ConfigurationT', bound=Configuration)


class FileBasedConfiguration(Configuration):
    def __init__(self):
        super().__init__()
        self._project_directory: TemporaryDirectory | None = None  # type: ignore[type-arg]
        self._configuration_file_path: Path | None = None
        self._autowrite = False

    @property
    def autowrite(self) -> bool:
        return self._autowrite

    @autowrite.setter
    def autowrite(self, autowrite: bool) -> None:
        if autowrite:
            if not self._autowrite:
                self.react.react_weakref(self._write_reactor)
        else:
            self.react.shutdown(self._write_reactor)
        self._autowrite = autowrite

    @sync
    async def _write_reactor(self) -> None:
        await self.write()

    async def write(self, configuration_file_path: Path | None = None) -> None:
        if configuration_file_path is not None:
            self.configuration_file_path = configuration_file_path

        await self._write(self.configuration_file_path)

    async def _write(self, configuration_file_path: Path) -> None:
        # Change the working directory to allow absolute paths to be turned relative to the configuration file's directory
        # path.
        formats = FormatRepository()
        async with ChDir(configuration_file_path.parent):
            dump = formats.format_for(configuration_file_path.suffix[1:]).dump(self.dump())
            try:
                async with aiofiles.open(configuration_file_path, mode='w') as f:
                    await f.write(dump)
            except FileNotFoundError:
                await makedirs(configuration_file_path.parent)
                await self.write()
        self._configuration_file_path = configuration_file_path

    async def read(self, configuration_file_path: Path | None = None) -> None:
        if configuration_file_path is not None:
            self.configuration_file_path = configuration_file_path

        formats = FormatRepository()
        with SerdeErrorCollection().assert_valid() as errors:
            # Change the working directory to allow relative paths to be resolved against the configuration file's directory
            # path.
            async with ChDir(self.configuration_file_path.parent):
                async with aiofiles.open(self.configuration_file_path) as f:
                    read_configuration = await f.read()
                with errors.catch(Str.plain(
                    'in {configuration_file_path}',
                    configuration_file_path=str(self.configuration_file_path.resolve()),
                )):
                    loaded_configuration = self.load(
                        formats.format_for(self.configuration_file_path.suffix[1:]).load(read_configuration)
                    )
        self.update(loaded_configuration)

    def __del__(self) -> None:
        if hasattr(self, '_project_directory') and self._project_directory is not None:
            self._project_directory.cleanup()

    @property
    @reactive_property
    def configuration_file_path(self) -> Path:
        if self._configuration_file_path is None:
            if self._project_directory is None:
                self._project_directory = TemporaryDirectory()
            wait(self._write(Path(self._project_directory.name) / f'{type(self).__name__}.json'))
        return cast(Path, self._configuration_file_path)

    @configuration_file_path.setter
    def configuration_file_path(self, configuration_file_path: Path) -> None:
        if configuration_file_path == self._configuration_file_path:
            return
        formats = FormatRepository()
        formats.format_for(configuration_file_path.suffix[1:])
        self._configuration_file_path = configuration_file_path

    @configuration_file_path.deleter
    def configuration_file_path(self) -> None:
        if self._autowrite:
            raise RuntimeError('Cannot remove the configuration file path while autowrite is enabled.')
        self._configuration_file_path = None


ConfigurationKey: TypeAlias = SupportsIndex | Hashable | type[Any]
ConfigurationKeyT = TypeVar('ConfigurationKeyT', bound=ConfigurationKey)


class ConfigurationCollection(Configuration, Generic[ConfigurationKeyT, ConfigurationT]):
    _configurations: MutableSequence[ConfigurationT] | MutableMapping[ConfigurationKeyT, ConfigurationT]

    def __init__(
        self,
        configurations: Iterable[ConfigurationT] | None = None,
    ):
        super().__init__()
        if configurations is not None:
            self.append(*configurations)

    @scope.register_self
    def __iter__(self) -> Iterator[ConfigurationKeyT] | Iterator[ConfigurationT]:
        raise NotImplementedError(repr(self))

    @scope.register_self
    def __contains__(self, item: Any) -> bool:
        return item in self._configurations

    def __getitem__(self, configuration_key: ConfigurationKeyT) -> ConfigurationT:
        raise NotImplementedError(repr(self))

    def __delitem__(self, configuration_key: ConfigurationKeyT) -> None:
        self.remove(configuration_key)

    @scope.register_self
    def __len__(self) -> int:
        return len(self._configurations)

    @scope.register_self
    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        scope.register(other)
        if list(self.keys()) != list(other.keys()):
            return False
        if list(self.values()) != list(other.values()):
            return False
        return True

    @recursive_repr()
    def __repr__(self) -> str:
        return repr_instance(self, configurations=list(self.values()))

    def _remove_without_trigger(self, *configuration_keys: ConfigurationKeyT) -> None:
        for configuration_key in configuration_keys:
            with suppress(LookupError):
                self._on_remove(self._configurations[configuration_key])  # type: ignore[call-overload]
            del self._configurations[configuration_key]  # type: ignore[call-overload]

    def remove(self, *configuration_keys: ConfigurationKeyT) -> None:
        self._remove_without_trigger(*configuration_keys)
        self.react.trigger()

    def _clear_without_trigger(self) -> None:
        self._remove_without_trigger(*self.keys())

    def clear(self) -> None:
        self._clear_without_trigger()
        self.react.trigger()

    def _on_add(self, configuration: ConfigurationT) -> None:
        configuration.react(self)

    def _on_remove(self, configuration: ConfigurationT) -> None:
        configuration.react.shutdown(self)

    def to_index(self, configuration_key: ConfigurationKeyT) -> int:
        raise NotImplementedError(repr(self))

    def to_indices(self, *configuration_keys: ConfigurationKeyT) -> Iterator[int]:
        for configuration_key in configuration_keys:
            yield self.to_index(configuration_key)

    def to_key(self, index: int) -> ConfigurationKeyT:
        raise NotImplementedError(repr(self))

    def to_keys(self, *indices: int | slice) -> Iterator[ConfigurationKeyT]:
        unique_indices = set()
        for index in indices:
            if isinstance(index, slice):
                for slice_index in slice_to_range(index, self._configurations):
                    unique_indices.add(slice_index)
            else:
                unique_indices.add(index)
        for index in sorted(unique_indices):
            yield self.to_key(index)

    @classmethod
    def _item_type(cls) -> type[ConfigurationT]:
        raise NotImplementedError(repr(cls))

    @classmethod
    def _create_default_item(cls, configuration_key: ConfigurationKeyT) -> ConfigurationT:
        raise NotImplementedError(repr(cls))

    def keys(self) -> Iterator[ConfigurationKeyT]:
        raise NotImplementedError(repr(self))

    def values(self) -> Iterator[ConfigurationT]:
        raise NotImplementedError(repr(self))

    def prepend(self, *configurations: ConfigurationT) -> None:
        raise NotImplementedError(repr(self))

    def append(self, *configurations: ConfigurationT) -> None:
        raise NotImplementedError(repr(self))

    def insert(self, index: int, *configurations: ConfigurationT) -> None:
        raise NotImplementedError(repr(self))

    def move_to_beginning(self, *configuration_keys: ConfigurationKeyT) -> None:
        raise NotImplementedError(repr(self))

    def move_towards_beginning(self, *configuration_keys: ConfigurationKeyT) -> None:
        raise NotImplementedError(repr(self))

    def move_to_end(self, *configuration_keys: ConfigurationKeyT) -> None:
        raise NotImplementedError(repr(self))

    def move_towards_end(self, *configuration_keys: ConfigurationKeyT) -> None:
        raise NotImplementedError(repr(self))


class ConfigurationSequence(ConfigurationCollection[int, ConfigurationT], Generic[ConfigurationT]):
    def __init__(
        self,
        configurations: Iterable[ConfigurationT] | None = None,
    ):
        self._configurations: MutableSequence[ConfigurationT] = []
        super().__init__(configurations)

    def to_index(self, configuration_key: int) -> int:
        return configuration_key

    def to_key(self, index: int) -> int:
        return index

    @overload
    def __getitem__(self, configuration_key: int) -> ConfigurationT:
        pass

    @overload
    def __getitem__(self, configuration_key: slice) -> Sequence[ConfigurationT]:
        pass

    def __getitem__(self, configuration_key: int | slice) -> ConfigurationT | Sequence[ConfigurationT]:
        return self._configurations[configuration_key]

    @scope.register_self
    def __iter__(self) -> Iterator[ConfigurationT]:
        return (configuration for configuration in self._configurations)

    @scope.register_self
    def keys(self) -> Iterator[int]:
        return iter(range(0, len(self._configurations)))

    @scope.register_self
    def values(self) -> Iterator[ConfigurationT]:
        yield from self._configurations

    def update(self, other: Self) -> None:
        raise NotImplementedError(repr(self))

    @classmethod
    def load(
            cls,
            dump: Dump,
            configuration: Self | None = None,
    ) -> Self:
        if configuration is None:
            configuration = cls()
        else:
            configuration._clear_without_trigger()
        asserter = Asserter()
        with SerdeErrorCollection().assert_valid():
            configuration.append(*asserter.assert_sequence(Assertions(cls._item_type().assert_load()))(dump))
        return configuration

    def dump(self) -> VoidableDump:
        return minimize([
            configuration.dump()
            for configuration in self._configurations
        ])

    def prepend(self, *configurations: ConfigurationT) -> None:
        for configuration in configurations:
            self._on_add(configuration)
            self._configurations.insert(0, configuration)
        self.react.trigger()

    def append(self, *configurations: ConfigurationT) -> None:
        for configuration in configurations:
            self._on_add(configuration)
            self._configurations.append(configuration)
        self.react.trigger()

    def insert(self, index: int, *configurations: ConfigurationT) -> None:
        for configuration in reversed(configurations):
            self._on_add(configuration)
            self._configurations.insert(index, configuration)
        self.react.trigger()

    def move_to_beginning(self, *configuration_keys: int) -> None:
        self.move_to_end(
            *configuration_keys,
            *[
                index
                for index
                in range(0, len(self._configurations))
                if index not in configuration_keys
            ]
        )

    def move_towards_beginning(self, *configuration_keys: int) -> None:
        for index in configuration_keys:
            self._configurations.insert(index - 1, self._configurations.pop(index))
        self.react.trigger()

    def move_to_end(self, *configuration_keys: int) -> None:
        for index in configuration_keys:
            self._configurations.append(self._configurations[index])
        for index in reversed(configuration_keys):
            self._configurations.pop(index)
        self.react.trigger()

    def move_towards_end(self, *configuration_keys: int) -> None:
        for index in reversed(configuration_keys):
            self._configurations.insert(index + 1, self._configurations.pop(index))
        self.react.trigger()


class ConfigurationMapping(ConfigurationCollection[ConfigurationKeyT, ConfigurationT], Generic[ConfigurationKeyT, ConfigurationT]):
    def __init__(
        self,
        configurations: Iterable[ConfigurationT] | None = None,
    ):
        self._configurations: OrderedDict[ConfigurationKeyT, ConfigurationT] = OrderedDict()
        super().__init__(configurations)

    def _minimize_item_dump(self) -> bool:
        return False

    def to_index(self, configuration_key: ConfigurationKeyT) -> int:
        return list(self._configurations.keys()).index(configuration_key)

    def to_key(self, index: int) -> ConfigurationKeyT:
        return list(self._configurations.keys())[index]

    @scope.register_self
    def __getitem__(self, configuration_key: ConfigurationKeyT) -> ConfigurationT:
        try:
            return self._configurations[configuration_key]
        except KeyError:
            self.append(self._create_default_item(configuration_key))
            return self._configurations[configuration_key]

    @scope.register_self
    def __iter__(self) -> Iterator[ConfigurationKeyT]:
        return (configuration_key for configuration_key in self._configurations)

    def _keys_without_scope(self) -> Iterator[ConfigurationKeyT]:
        return (configuration_key for configuration_key in self._configurations.keys())

    @scope.register_self
    def keys(self) -> Iterator[ConfigurationKeyT]:
        return self._keys_without_scope()

    @scope.register_self
    def values(self) -> Iterator[ConfigurationT]:
        yield from self._configurations.values()

    def update(self, other: Self) -> None:
        self.replace(*other.values())

    def replace(self, *values: ConfigurationT) -> None:
        self_keys = OrderedSet(self.keys())
        other = {
            self._get_key(value): value
            for value in values
        }
        other_values = list(values)
        other_keys = OrderedSet(map(self._get_key, other_values))

        # Update items that are kept.
        for key in self_keys & other_keys:
            self[key].update(other[key])

        # Add items that are new.
        self._append_without_trigger(*(other[key] for key in (other_keys - self_keys)))

        # Remove items that should no longer be present.
        self._remove_without_trigger(*(self_keys - other_keys))

        # Ensure everything is in the correct order. This will also trigger reactors.
        self.move_to_beginning(*other_keys)

    @classmethod
    def load(
            cls,
            dump: Dump,
            configuration: Self | None = None,
    ) -> Self:
        if configuration is None:
            configuration = cls()
        asserter = Asserter()
        dict_dump = asserter.assert_dict()(dump)
        mapping = asserter.assert_mapping(Assertions(cls._item_type().load))({
            key: cls._load_key(value, key)
            for key, value
            in dict_dump.items()
        })
        configuration.replace(*mapping.values())
        return configuration

    def dump(self) -> VoidableDump:
        dump = {}
        for configuration_item in self._configurations.values():
            item_dump = configuration_item.dump()
            if item_dump is not Void:
                item_dump, configuration_key = self._dump_key(item_dump)
                if self._minimize_item_dump():
                    item_dump = minimize(item_dump)
                dump[configuration_key] = item_dump
        return minimize(dump)

    def prepend(self, *configurations: ConfigurationT) -> None:
        for configuration in configurations:
            configuration_key = self._get_key(configuration)
            self._configurations[configuration_key] = configuration
            configuration.react(self)
        self.move_to_beginning(*map(self._get_key, configurations))

    def _append_without_trigger(self, *configurations: ConfigurationT) -> None:
        for configuration in configurations:
            configuration_key = self._get_key(configuration)
            self._configurations[configuration_key] = configuration
            configuration.react(self)
        self._move_to_end_without_trigger(*map(self._get_key, configurations))

    def append(self, *configurations: ConfigurationT) -> None:
        self._append_without_trigger(*configurations)
        self.react.trigger()

    def _insert_without_trigger(self, index: int, *configurations: ConfigurationT) -> None:
        current_configuration_keys = list(self._keys_without_scope())
        self._append_without_trigger(*configurations)
        self._move_to_end_without_trigger(
            *current_configuration_keys[0:index],
            *map(self._get_key, configurations),
            *current_configuration_keys[index:]
        )

    def insert(self, index: int, *configurations: ConfigurationT) -> None:
        self._insert_without_trigger(index, *configurations)
        self.react.trigger()

    def move_to_beginning(self, *configuration_keys: ConfigurationKeyT) -> None:
        for configuration_key in reversed(configuration_keys):
            self._configurations.move_to_end(configuration_key, False)
        self.react.trigger()

    def move_towards_beginning(self, *configuration_keys: ConfigurationKeyT) -> None:
        self._move_by_offset(-1, *configuration_keys)

    def _move_to_end_without_trigger(self, *configuration_keys: ConfigurationKeyT) -> None:
        for configuration_key in configuration_keys:
            self._configurations.move_to_end(configuration_key)

    def move_to_end(self, *configuration_keys: ConfigurationKeyT) -> None:
        self._move_to_end_without_trigger(*configuration_keys)
        self.react.trigger()

    def move_towards_end(self, *configuration_keys: ConfigurationKeyT) -> None:
        self._move_by_offset(1, *configuration_keys)

    def _move_by_offset(self, offset: int, *configuration_keys: ConfigurationKeyT) -> None:
        current_configuration_keys = list(self._keys_without_scope())
        indices = list(self.to_indices(*configuration_keys))
        if offset > 0:
            indices.reverse()
        for index in indices:
            self._insert_without_trigger(index + offset, self._configurations.pop(current_configuration_keys[index]))
        self.react.trigger()

    def _get_key(self, configuration: ConfigurationT) -> ConfigurationKeyT:
        raise NotImplementedError(repr(self))

    @classmethod
    def _load_key(
        cls,
        item_dump: Dump,
        key_dump: str,
    ) -> Dump:
        raise NotImplementedError(repr(cls))

    def _dump_key(self, item_dump: VoidableDump) -> tuple[VoidableDump, str]:
        raise NotImplementedError(repr(self))


class Configurable(Generic[ConfigurationT]):
    _configuration: ConfigurationT

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)

    @property
    def configuration(self) -> ConfigurationT:
        if not hasattr(self, '_configuration'):
            raise RuntimeError(f'{self} has no configuration. {type(self)}.__init__() must ensure it is set.')
        return self._configuration



from __future__ import annotations

import hashlib
import logging
import os
import shutil
import sys
from asyncio import subprocess as aiosubprocess
from contextlib import suppress
from pathlib import Path
from subprocess import CalledProcessError
from typing import Sequence, Any

from aiofiles.tempfile import TemporaryDirectory

from betty import subprocess
from betty.app.extension import Extension, discover_extension_types
from betty.app.extension.requirement import Requirement, AnyRequirement, AllRequirements
from betty.asyncio import sync
from betty.cache import CacheScope
from betty.fs import iterfiles
from betty.locale import Str


async def npm(arguments: Sequence[str], **kwargs: Any) -> aiosubprocess.Process:
    # Use a shell on Windows so subprocess can find the executables it needs (see
    # https://bugs.python.org/issue17023).
    runner = subprocess.run_shell if sys.platform.startswith('win32') else subprocess.run_exec
    return await runner(['npm', *arguments], **kwargs)


class _NpmRequirement(Requirement):
    def __init__(self, met: bool):
        super().__init__()
        self._met = met
        self._summary = self._met_summary() if met else self._unmet_summary()
        self._details = Str._('npm (https://www.npmjs.com/) must be available for features that require Node.js packages to be installed. Ensure that the `npm` executable is available in your `PATH`.')

    @classmethod
    def _met_summary(cls) -> Str:
        return Str._('`npm` is available')

    @classmethod
    def _unmet_summary(cls) -> Str:
        return Str._('`npm` is not available')

    @classmethod
    @sync
    async def check(cls) -> _NpmRequirement:
        try:
            await npm(['--version'])
            logging.getLogger().debug(cls._met_summary())
            return cls(True)
        except (CalledProcessError, FileNotFoundError):
            logging.getLogger().debug(cls._unmet_summary())
            return cls(False)

    def is_met(self) -> bool:
        return self._met

    def summary(self) -> Str:
        return self._summary

    def details(self) -> Str:
        return self._details


def is_assets_build_directory_path(path: Path) -> bool:
    return path.is_dir() and len(os.listdir(path)) > 0


class _AssetsRequirement(Requirement):
    def __init__(self, extension_types: set[type[NpmBuilder & Extension]]):
        super().__init__()
        self._extension_types = extension_types
        self._summary = Str._('Pre-built assets')
        self._details: Str
        if not self.is_met():
            extension_names = sorted(
                extension_type.name()
                for extension_type
                in self._extension_types - self._extension_types_with_built_assets
            )
            self._details = Str._(
                'Pre-built assets are unavailable for {extension_names}.',
                extension_names=', '.join(extension_names,
                                          ))
        else:
            self._details = Str.plain('')

    @property
    def _extension_types_with_built_assets(self) -> set[type[NpmBuilder & Extension]]:
        return {
            extension_type
            for extension_type
            in self._extension_types
            if is_assets_build_directory_path(_get_assets_build_directory_path(extension_type))
        }

    def is_met(self) -> bool:
        return self._extension_types <= self._extension_types_with_built_assets

    def summary(self) -> Str:
        return self._summary

    def details(self) -> Str:
        return self._details


class NpmBuilder:
    async def npm_build(self, working_directory_path: Path, assets_directory_path: Path) -> None:
        raise NotImplementedError(repr(self))

    @classmethod
    def npm_cache_scope(cls) -> CacheScope:
        return CacheScope.PROJECT


def discover_npm_builders() -> set[type[NpmBuilder & Extension]]:
    return {
        extension_type
        for extension_type
        in discover_extension_types()
        if issubclass(extension_type, NpmBuilder)
    }


def _get_assets_directory_path(extension_type: type[NpmBuilder & Extension]) -> Path:
    assert issubclass(extension_type, Extension)
    assert issubclass(extension_type, NpmBuilder)
    assets_directory_path = extension_type.assets_directory_path()
    if not assets_directory_path:
        raise RuntimeError(f'Extension {extension_type} does not have an assets directory.')
    return assets_directory_path / _Npm.name()


def _get_assets_src_directory_path(extension_type: type[NpmBuilder & Extension]) -> Path:
    return _get_assets_directory_path(extension_type) / 'src'


def _get_assets_build_directory_path(extension_type: type[NpmBuilder & Extension]) -> Path:
    return _get_assets_directory_path(extension_type) / 'build'


async def build_assets(extension: NpmBuilder & Extension) -> Path:
    assets_directory_path = _get_assets_build_directory_path(type(extension))
    await _build_assets_to_directory_path(extension, assets_directory_path)
    return assets_directory_path


async def _build_assets_to_directory_path(extension: NpmBuilder & Extension, assets_directory_path: Path) -> None:
    assert isinstance(extension, Extension)
    assert isinstance(extension, NpmBuilder)
    with suppress(FileNotFoundError):
        shutil.rmtree(assets_directory_path)
    os.makedirs(assets_directory_path)
    async with TemporaryDirectory() as working_directory_path_str:
        working_directory_path = Path(working_directory_path_str)
        await extension.npm_build(Path(working_directory_path), assets_directory_path)


class _Npm(Extension):
    _npm_requirement: _NpmRequirement | None = None
    _assets_requirement: _AssetsRequirement | None = None
    _requirement: Requirement | None = None

    @classmethod
    def _ensure_requirement(cls) -> Requirement:
        if cls._requirement is None:
            cls._npm_requirement = _NpmRequirement.check()
            cls._assets_requirement = _AssetsRequirement(discover_npm_builders())
            assert cls._npm_requirement is not None
            assert cls._assets_requirement is not None
            cls._requirement = AnyRequirement(cls._npm_requirement, cls._assets_requirement)
        return cls._requirement

    @classmethod
    def enable_requirement(cls) -> Requirement:
        return AllRequirements(
            cls._ensure_requirement(),
            super().enable_requirement(),
        )

    async def install(self, extension_type: type[NpmBuilder & Extension], working_directory_path: Path) -> None:
        self._ensure_requirement()
        if self._npm_requirement:
            self._npm_requirement.assert_met()

        shutil.copytree(
            _get_assets_src_directory_path(extension_type),
            working_directory_path,
            dirs_exist_ok=True,
        )
        async for file_path in iterfiles(working_directory_path):
            await self._app.renderer.render_file(file_path)
        await npm(['install', '--production'], cwd=working_directory_path)

    def _get_cached_assets_build_directory_path(self, extension_type: type[NpmBuilder & Extension]) -> Path:
        path = self.cache_directory_path / extension_type.name()
        if extension_type.npm_cache_scope() == CacheScope.PROJECT:
            path /= hashlib.md5(str(self.app.project.configuration.configuration_file_path).encode('utf-8')).hexdigest()
        return path

    async def ensure_assets(self, extension: NpmBuilder & Extension) -> Path:
        assets_build_directory_paths = [
            _get_assets_build_directory_path(type(extension)),
            self._get_cached_assets_build_directory_path(type(extension)),
        ]
        for assets_build_directory_path in assets_build_directory_paths:
            if is_assets_build_directory_path(assets_build_directory_path):
                return assets_build_directory_path

        if self._npm_requirement:
            self._npm_requirement.assert_met()
        return await self._build_cached_assets(extension)

    async def _build_cached_assets(self, extension: NpmBuilder & Extension) -> Path:
        assets_directory_path = self._get_cached_assets_build_directory_path(type(extension))
        await _build_assets_to_directory_path(extension, assets_directory_path)
        return assets_directory_path



import logging
import shutil
from contextlib import suppress
from enum import unique, IntFlag, auto

from betty import fs
from betty.locale import Localizer


@unique
class CacheScope(IntFlag):
    BETTY = auto()
    PROJECT = auto()


class Cache:
    def __init__(self, localizer: Localizer):
        self._localizer = localizer

    async def clear(self) -> None:
        with suppress(FileNotFoundError):
            shutil.rmtree(fs.CACHE_DIRECTORY_PATH)
        logging.getLogger().info(self._localizer._('All caches cleared.'))



from __future__ import annotations

from contextlib import suppress
from enum import Enum
from functools import total_ordering
from pathlib import Path
from reprlib import recursive_repr
from typing import Iterable, Any

from geopy import Point

from betty.classtools import repr_instance
from betty.locale import Localized, Datey, Str, Localizable
from betty.media_type import MediaType
from betty.model import many_to_many, Entity, one_to_many, many_to_one, many_to_one_to_many, \
    MultipleTypesEntityCollection, EntityCollection, UserFacingEntity, EntityTypeAssociationRegistry, \
    PickleableEntityGraph
from betty.model.event_type import EventType, StartOfLifeEventType, EndOfLifeEventType
from betty.pickle import State, Pickleable


class Privacy(Enum):
    PUBLIC = 1
    PRIVATE = 2
    UNDETERMINED = 3


class HasPrivacy(Pickleable):
    def __init__(
        self,
        *args: Any,
        privacy: Privacy | None = None,
        public: bool | None = None,
        private: bool | None = None,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        if [privacy, public, private].count(None) < 2:
            raise ValueError(f'Only one of the `privacy`, `public`, and `private` arguments to {type(self)}.__init__() may be given at a time.')
        if privacy is not None:
            self._privacy = privacy
        elif public is True:
            self._privacy = Privacy.PUBLIC
        elif private is True:
            self._privacy = Privacy.PRIVATE
        else:
            self._privacy = Privacy.UNDETERMINED

    def __getstate__(self) -> State:
        dict_state, slots_state = super().__getstate__()
        dict_state['_privacy'] = self._privacy
        return dict_state, slots_state

    def _get_privacy(self) -> Privacy:
        return self._privacy

    @property
    def privacy(self) -> Privacy:
        return self._get_privacy()

    @property
    def private(self) -> bool:
        return self._get_privacy() is Privacy.PRIVATE

    @property
    def public(self) -> bool:
        # Undetermined privacy defaults to public.
        return self._get_privacy() is not Privacy.PRIVATE


class HasMutablePrivacy(HasPrivacy):
    @property
    def privacy(self) -> Privacy:
        return self._get_privacy()

    @privacy.setter
    def privacy(self, privacy: Privacy) -> None:
        self._privacy = privacy

    @privacy.deleter
    def privacy(self) -> None:
        self._privacy = Privacy.UNDETERMINED

    @property
    def private(self) -> bool:
        return self._get_privacy() is Privacy.PRIVATE

    @private.setter
    def private(self, private: True) -> None:
        self._privacy = Privacy.PRIVATE

    @property
    def public(self) -> bool:
        # Undetermined privacy defaults to public.
        return self._get_privacy() is not Privacy.PRIVATE

    @public.setter
    def public(self, public: True) -> None:
        self._privacy = Privacy.PUBLIC


def is_private(target: Any) -> bool:
    if isinstance(target, HasPrivacy):
        return target.private
    return False


def is_public(target: Any) -> bool:
    if isinstance(target, HasPrivacy):
        return target.public
    return True


def resolve_privacy(privacy: Privacy | HasPrivacy | None) -> Privacy:
    if privacy is None:
        return Privacy.UNDETERMINED
    if isinstance(privacy, Privacy):
        return privacy
    return privacy.privacy


def merge_privacies(*privacies: Privacy | HasPrivacy | None) -> Privacy:
    privacies = {
        resolve_privacy(privacy)
        for privacy
        in privacies
    }
    if Privacy.PRIVATE in privacies:
        return Privacy.PRIVATE
    if Privacy.UNDETERMINED in privacies:
        return Privacy.UNDETERMINED
    return Privacy.PUBLIC


class Dated(Pickleable):
    def __init__(
        self,
        *args: Any,
        date: Datey | None = None,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        self.date = date

    def __getstate__(self) -> State:
        dict_state, slots_state = super().__getstate__()
        dict_state['date'] = self.date
        return dict_state, slots_state


@many_to_one('entity', 'betty.model.ancestry.HasNotes', 'notes')
class Note(HasMutablePrivacy, UserFacingEntity, Entity):
    entity: HasNotes

    def __init__(
        self,
        text: str,
        *,
        id: str | None = None,
        entity: HasNotes | None = None,
        privacy: Privacy | None = None,
        public: bool | None = None,
        private: bool | None = None,
    ):
        super().__init__(
            id,
            privacy=privacy,
            public=public,
            private=private,
        )
        self._text = text
        if entity is not None:
            self.entity = entity

    def __getstate__(self) -> State:
        dict_state, slots_state = super().__getstate__()
        dict_state['_text'] = self._text
        return dict_state, slots_state

    @classmethod
    def entity_type_label(cls) -> Str:
        return Str._('Note')

    @classmethod
    def entity_type_label_plural(cls) -> Str:
        return Str._('Notes')

    @property
    def text(self) -> str:
        return self._text

    @property
    def label(self) -> Str:
        return Str.plain(self.text)


@one_to_many('notes', 'betty.model.ancestry.Note', 'entity')
class HasNotes:
    def __init__(  # type: ignore[misc]
        self: HasNotes & Entity,
        *args: Any,
        notes: Iterable[Note] | None = None,
        **kwargs: Any,
    ):
        super().__init__(  # type: ignore[misc]
            *args,
            **kwargs,
        )
        if notes is not None:
            self.notes = notes  # type: ignore[assignment]

    @property
    def notes(self) -> EntityCollection[Note]:  # type: ignore[empty-body]
        pass

    @notes.setter
    def notes(self, notes: Iterable[Note]) -> None:
        pass

    @notes.deleter
    def notes(self) -> None:
        pass


class Described(Pickleable):
    def __init__(
        self,
        *args: Any,
        description: str | None = None,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        self.description = description

    def __getstate__(self) -> State:
        dict_state, slots_state = super().__getstate__()
        dict_state['description'] = self.description
        return dict_state, slots_state


class HasMediaType(Pickleable):
    def __init__(
        self,
        *args: Any,
        media_type: MediaType | None = None,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        self.media_type = media_type

    def __getstate__(self) -> State:
        dict_state, slots_state = super().__getstate__()
        dict_state['media_type'] = self.media_type
        return dict_state, slots_state


class Link(HasMediaType, Localized, Described):
    url: str
    relationship: str | None
    label: str | None

    def __init__(
        self,
        url: str,
        *,
        relationship: str | None = None,
        label: str | None = None,
        description: str | None = None,
        media_type: MediaType | None = None,
        locale: str | None = None,
    ):
        super().__init__(
            media_type=media_type,
            description=description,
            locale=locale,
        )
        self.url = url
        self.label = label
        self.relationship = relationship

    def __getstate__(self) -> State:
        dict_state, slots_state = super().__getstate__()
        dict_state['url'] = self.url
        dict_state['relationship'] = self.relationship
        dict_state['label'] = self.label
        return dict_state, slots_state


class HasLinks(Pickleable):
    def __init__(
        self,
        *args: Any,
        links: set[Link] | None = None,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        self._links: set[Link] = set() if links is None else links

    def __getstate__(self) -> State:
        dict_state, slots_state = super().__getstate__()
        dict_state['_links'] = self._links
        return dict_state, slots_state

    @property
    def links(self) -> set[Link]:
        return self._links


@many_to_many('citations', 'betty.model.ancestry.Citation', 'facts')
class HasCitations:
    def __init__(  # type: ignore[misc]
        self: HasCitations & Entity,
        *args: Any,
        citations: Iterable[Citation] | None = None,
        **kwargs: Any,
    ):
        super().__init__(  # type: ignore[misc]
            *args,
            **kwargs,
        )
        if citations is not None:
            self.citations = citations  # type: ignore[assignment]

    @property
    def citations(self) -> EntityCollection[Citation]:  # type: ignore[empty-body]
        pass

    @citations.setter
    def citations(self, citations: Iterable[Citation]) -> None:
        pass

    @citations.deleter
    def citations(self) -> None:
        pass


@many_to_many('entities', 'betty.model.ancestry.HasFiles', 'files')
class File(Described, HasMutablePrivacy, HasMediaType, HasNotes, HasCitations, UserFacingEntity, Entity):
    def __init__(
        self,
        path: Path,
        *,
        id: str | None = None,
        media_type: MediaType | None = None,
        description: str | None = None,
        notes: Iterable[Note] | None = None,
        citations: Iterable[Citation] | None = None,
        privacy: Privacy | None = None,
        public: bool | None = None,
        private: bool | None = None,
    ):
        super().__init__(
            id,
            media_type=media_type,
            description=description,
            notes=notes,
            citations=citations,
            privacy=privacy,
            public=public,
            private=private,
        )
        self._path = path

    def __getstate__(self) -> State:
        dict_state, slots_state = super().__getstate__()
        dict_state['_path'] = self._path
        return dict_state, slots_state

    @property
    def entities(self) -> EntityCollection[Entity]:  # type: ignore[empty-body]
        pass

    @entities.setter
    def entities(self, entities: Iterable[Entity]) -> None:
        pass

    @entities.deleter
    def entities(self) -> None:
        pass

    @classmethod
    def entity_type_label(cls) -> Str:
        return Str._('File')

    @classmethod
    def entity_type_label_plural(cls) -> Str:
        return Str._('Files')

    @property
    def path(self) -> Path:
        return self._path

    @property
    def label(self) -> Str:
        return Str.plain(self.description) if self.description else super().label


@many_to_many('files', 'betty.model.ancestry.File', 'entities')
class HasFiles:
    def __init__(  # type: ignore[misc]
        self: HasFiles & Entity,
        *args: Any,
        files: Iterable[File] | None = None,
        **kwargs: Any,
    ):
        super().__init__(  # type: ignore[misc]
            *args,
            **kwargs,
        )
        if files is not None:
            self.files = files  # type: ignore[assignment]

    @property
    def files(self) -> EntityCollection[File]:  # type: ignore[empty-body]
        pass

    @files.setter
    def files(self, files: Iterable[File]) -> None:
        pass

    @files.deleter
    def files(self) -> None:
        pass

    @property
    def associated_files(self) -> Iterable[File]:
        return self.files


@many_to_one('contained_by', 'betty.model.ancestry.Source', 'contains')
@one_to_many('contains', 'betty.model.ancestry.Source', 'contained_by')
@one_to_many('citations', 'betty.model.ancestry.Citation', 'source')
class Source(Dated, HasFiles, HasLinks, HasMutablePrivacy, UserFacingEntity, Entity):
    contained_by: Source | None

    def __init__(
        self,
        name: str | None = None,
        *,
        id: str | None = None,
        author: str | None = None,
        publisher: str | None = None,
        contained_by: Source | None = None,
        contains: Iterable[Source] | None = None,
        date: Datey | None = None,
        files: Iterable[File] | None = None,
        links: set[Link] | None = None,
        privacy: Privacy | None = None,
        public: bool | None = None,
        private: bool | None = None,
    ):
        super().__init__(
            id,
            date=date,
            files=files,
            links=links,
            privacy=privacy,
            public=public,
            private=private,
        )
        self.name = name
        self.author = author
        self.publisher = publisher
        if contained_by is not None:
            self.contained_by = contained_by
        if contains is not None:
            self.contains = contains  # type: ignore[assignment]

    def __getstate__(self) -> State:
        dict_state, slots_state = super().__getstate__()
        dict_state['name'] = self.name
        dict_state['author'] = self.author
        dict_state['publisher'] = self.publisher
        return dict_state, slots_state

    def _get_privacy(self) -> Privacy:
        privacy = super()._get_privacy()
        if self.contained_by:
            return merge_privacies(privacy, self.contained_by.privacy)
        return privacy

    @property
    def contains(self) -> EntityCollection[Source]:  # type: ignore[empty-body]
        pass

    @contains.setter
    def contains(self, contains: Iterable[Source]) -> None:
        pass

    @contains.deleter
    def contains(self) -> None:
        pass

    @property
    def citations(self) -> EntityCollection[Citation]:  # type: ignore[empty-body]
        pass

    @citations.setter
    def citations(self, citations: Iterable[Citation]) -> None:
        pass

    @citations.deleter
    def citations(self) -> None:
        pass

    @classmethod
    def entity_type_label(cls) -> Str:
        return Str._('Source')

    @classmethod
    def entity_type_label_plural(cls) -> Str:
        return Str._('Sources')

    @property
    def label(self) -> Str:
        return Str.plain(self.name) if self.name else super().label


class AnonymousSource(Source):
    @property  # type: ignore[override]
    def name(self) -> str:
        return 'private'

    @name.setter
    def name(self, _) -> None:
        # This is a no-op as the name is 'hardcoded'.
        pass

    @name.deleter
    def name(self) -> None:
        # This is a no-op as the name is 'hardcoded'.
        pass


@many_to_many('facts', 'betty.model.ancestry.HasCitations', 'citations')
@many_to_one('source', 'betty.model.ancestry.Source', 'citations')
class Citation(Dated, HasFiles, HasMutablePrivacy, UserFacingEntity, Entity):
    def __init__(
        self,
        *,
        id: str | None = None,
        facts: Iterable[HasCitations] | None = None,
        source: Source | None = None,
        location: Str | None = None,
        date: Datey | None = None,
        files: Iterable[File] | None = None,
        privacy: Privacy | None = None,
        public: bool | None = None,
        private: bool | None = None,
    ):
        super().__init__(
            id,
            date=date,
            files=files,
            privacy=privacy,
            public=public,
            private=private,
        )
        if facts is not None:
            self.facts = facts  # type: ignore[assignment]
        self.location = location
        self.source = source

    def __getstate__(self) -> State:
        dict_state, slots_state = super().__getstate__()
        dict_state['location'] = self.location
        return dict_state, slots_state

    def _get_privacy(self) -> Privacy:
        privacy = super()._get_privacy()
        if self.source:
            return merge_privacies(privacy, self.source.privacy)
        return privacy

    @property
    def facts(self) -> EntityCollection[HasCitations & Entity]:  # type: ignore[empty-body]
        pass

    @facts.setter
    def facts(self, facts: Iterable[HasCitations & Entity]) -> None:
        pass

    @facts.deleter
    def facts(self) -> None:
        pass

    @classmethod
    def entity_type_label(cls) -> Str:
        return Str._('Citation')

    @classmethod
    def entity_type_label_plural(cls) -> Str:
        return Str._('Citations')

    @property
    def label(self) -> Str:
        return self.location or Str.plain('')


class AnonymousCitation(Citation):
    @property  # type: ignore[override]
    def location(self) -> Str:
        return Str._("private (in order to protect people's privacy)")

    @location.setter
    def location(self, _) -> None:
        # This is a no-op as the location is 'hardcoded'.
        pass

    @location.deleter
    def location(self) -> None:
        # This is a no-op as the location is 'hardcoded'.
        pass


class PlaceName(Localized, Dated):
    def __init__(
        self,
        name: str,
        *,
        locale: str | None = None,
        date: Datey | None = None,
    ):
        super().__init__(
            date=date,
            locale=locale,
        )
        self._name = name

    def __getstate__(self) -> State:
        dict_state, slots_state = super().__getstate__()
        dict_state['_name'] = self._name
        return dict_state, slots_state

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented  # pragma: no cover
        return self._name == other._name and self.locale == other.locale

    @recursive_repr()
    def __repr__(self) -> str:
        return repr_instance(self, name=self.name, locale=self.locale)

    def __str__(self) -> str:
        return self._name

    @property
    def name(self) -> str:
        return self._name


@many_to_one_to_many(
    'betty.model.ancestry.Place',
    'enclosed_by',
    'encloses',
    'enclosed_by',
    'betty.model.ancestry.Place',
    'encloses',
)
class Enclosure(Dated, HasCitations, Entity):
    encloses: Place | None
    enclosed_by: Place | None

    def __init__(
        self,
        encloses: Place | None,
        enclosed_by: Place | None,
    ):
        super().__init__()
        self.encloses = encloses
        self.enclosed_by = enclosed_by

    @classmethod
    def entity_type_label(cls) -> Str:
        return Str._('Enclosure')

    @classmethod
    def entity_type_label_plural(cls) -> Str:
        return Str._('Enclosures')


@one_to_many('events', 'betty.model.ancestry.Event', 'place')
@one_to_many('enclosed_by', 'betty.model.ancestry.Enclosure', 'encloses')
@one_to_many('encloses', 'betty.model.ancestry.Enclosure', 'enclosed_by')
class Place(HasLinks, UserFacingEntity, Entity):
    def __init__(
        self,
        *,
        id: str | None = None,
        names: list[PlaceName] | None = None,
        events: Iterable[Event] | None = None,
        coordinates: Point | None = None,
        links: set[Link] | None = None,
    ):
        super().__init__(
            id,
            links=links,
        )
        self._names = [] if names is None else names
        self._coordinates = coordinates
        if events is not None:
            self.events = events  # type: ignore[assignment]

    def __getstate__(self) -> State:
        dict_state, slots_state = super().__getstate__()
        dict_state['_names'] = self._names
        dict_state['_coordinates'] = self._coordinates
        return dict_state, slots_state

    @property
    def enclosed_by(self) -> EntityCollection[Enclosure]:  # type: ignore[empty-body]
        pass

    @enclosed_by.setter
    def enclosed_by(self, enclosed_by: Iterable[Enclosure]) -> None:
        pass

    @enclosed_by.deleter
    def enclosed_by(self) -> None:
        pass

    @property
    def encloses(self) -> EntityCollection[Enclosure]:  # type: ignore[empty-body]
        pass

    @encloses.setter
    def encloses(self, encloses: Iterable[Enclosure]) -> None:
        pass

    @encloses.deleter
    def encloses(self) -> None:
        pass

    @property
    def events(self) -> EntityCollection[Event]:  # type: ignore[empty-body]
        pass

    @events.setter
    def events(self, events: Iterable[Event]) -> None:
        pass

    @events.deleter
    def events(self) -> None:
        pass

    @classmethod
    def entity_type_label(cls) -> Str:
        return Str._('Place')

    @classmethod
    def entity_type_label_plural(cls) -> Str:
        return Str._('Places')

    @property
    def names(self) -> list[PlaceName]:
        return self._names

    @property
    def coordinates(self) -> Point | None:
        return self._coordinates

    @coordinates.setter
    def coordinates(self, coordinates: Point):
        self._coordinates = coordinates

    @property
    def label(self) -> Str:
        # @todo Negotiate this by locale and date.
        with suppress(IndexError):
            return Str.plain(self.names[0].name)
        return super().label


class PresenceRole:
    @classmethod
    def name(cls) -> str:
        raise NotImplementedError(repr(cls))

    @property
    def label(self) -> Str:
        raise NotImplementedError(repr(self))


class Subject(PresenceRole):
    @classmethod
    def name(cls) -> str:
        return 'subject'

    @property
    def label(self) -> Str:
        return Str._('Subject')


class Witness(PresenceRole):
    @classmethod
    def name(cls) -> str:
        return 'witness'

    @property
    def label(self) -> Str:
        return Str._('Witness')


class Beneficiary(PresenceRole):
    @classmethod
    def name(cls) -> str:
        return 'beneficiary'

    @property
    def label(self) -> Str:
        return Str._('Beneficiary')


class Attendee(PresenceRole):
    @classmethod
    def name(cls) -> str:
        return 'attendee'

    @property
    def label(self) -> Str:
        return Str._('Attendee')


class Speaker(PresenceRole):
    @classmethod
    def name(cls) -> str:
        return 'speaker'

    @property
    def label(self) -> Str:
        return Str._('Speaker')


class Celebrant(PresenceRole):
    @classmethod
    def name(cls) -> str:
        return 'celebrant'

    @property
    def label(self) -> Str:
        return Str._('Celebrant')


class Organizer(PresenceRole):
    @classmethod
    def name(cls) -> str:
        return 'organizer'

    @property
    def label(self) -> Str:
        return Str._('Organizer')


@many_to_one_to_many(
    'betty.model.ancestry.Person',
    'presences',
    'person',
    'event',
    'betty.model.ancestry.Event',
    'presences',
)
class Presence(HasPrivacy, Entity):
    person: Person | None
    event: Event | None
    role: PresenceRole

    def __init__(
        self,
        person: Person | None,
        role: PresenceRole,
        event: Event | None,
    ):
        super().__init__(None)
        self.person = person
        self.role = role
        self.event = event

    def __getstate__(self) -> State:
        dict_state, slots_state = super().__getstate__()
        dict_state['role'] = self.role
        return dict_state, slots_state

    @classmethod
    def entity_type_label(cls) -> Str:
        return Str._('Presence')

    @classmethod
    def entity_type_label_plural(cls) -> Str:
        return Str._('Presences')

    def _get_privacy(self) -> Privacy:
        return merge_privacies(self.person, self.event)


@many_to_one('place', 'betty.model.ancestry.Place', 'events')
@one_to_many('presences', 'betty.model.ancestry.Presence', 'event')
class Event(Dated, HasFiles, HasCitations, Described, HasMutablePrivacy, UserFacingEntity, Entity):
    place: Place | None

    def __init__(
        self,
        *,
        id: str | None = None,
        event_type: type[EventType],
        date: Datey | None = None,
        files: Iterable[File] | None = None,
        citations: Iterable[Citation] | None = None,
        privacy: Privacy | None = None,
        public: bool | None = None,
        private: bool | None = None,
        place: Place | None = None,
        description: str | None = None,
    ):
        super().__init__(
            id,
            date=date,
            files=files,
            citations=citations,
            privacy=privacy,
            public=public,
            private=private,
            description=description,
        )
        self._event_type = event_type
        if place is not None:
            self.place = place

    @property
    def label(self) -> Str:
        format_kwargs: dict[str, str | Localizable] = {
            'event_type': self._event_type.label(),
        }
        subjects = [
            presence.person
            for presence
            in self.presences
            if presence.public and isinstance(presence.role, Subject) and presence.person is not None and presence.person.public
        ]
        if subjects:
            format_kwargs['subjects'] = Str.call(lambda localizer: ', '.join(person.label.localize(localizer) for person in subjects))
        if self.description is not None:
            format_kwargs['event_description'] = self.description

        if subjects:
            if self.description is None:
                return Str._('{event_type} of {subjects}', **format_kwargs)
            else:
                return Str._('{event_type} ({event_description}) of {subjects}', **format_kwargs)
        if self.description is None:
            return Str._('{event_type}', **format_kwargs)
        else:
            return Str._('{event_type} ({event_description})', **format_kwargs)

    def __getstate__(self) -> State:
        dict_state, slots_state = super().__getstate__()
        dict_state['_event_type'] = self._event_type
        return dict_state, slots_state

    @recursive_repr()
    def __repr__(self) -> str:
        return repr_instance(self, id=self._id, type=self._event_type)

    @property
    def presences(self) -> EntityCollection[Presence]:  # type: ignore[empty-body]
        pass

    @presences.setter
    def presences(self, presences: Iterable[Presence]) -> None:
        pass

    @presences.deleter
    def presences(self) -> None:
        pass

    @classmethod
    def entity_type_label(cls) -> Str:
        return Str._('Event')

    @classmethod
    def entity_type_label_plural(cls) -> Str:
        return Str._('Events')

    @property
    def event_type(self) -> type[EventType]:
        return self._event_type

    @property
    def associated_files(self) -> Iterable[File]:
        files = [
            *self.files,
            *[file for citation in self.citations for file in citation.associated_files],
        ]
        # Preserve the original order.
        seen = set()
        for file in files:
            if file in seen:
                continue
            seen.add(file)
            yield file


@total_ordering
@many_to_one('person', 'betty.model.ancestry.Person', 'names')
class PersonName(Localized, HasCitations, HasMutablePrivacy, Entity):
    person: Person | None

    def __init__(
        self,
        *,
        id: str | None = None,
        person: Person | None = None,
        individual: str | None = None,
        affiliation: str | None = None,
        privacy: Privacy | None = None,
        public: bool | None = None,
        private: bool | None = None,
    ):
        if not individual and not affiliation:
            raise ValueError('The individual and affiliation names must not both be empty.')
        super().__init__(
            id,
            privacy=privacy,
            public=public,
            private=private,
        )
        self._individual = individual
        self._affiliation = affiliation
        # Set the person association last, because the association requires comparisons, and self.__eq__() uses the
        # individual and affiliation names.
        self.person = person

    def __getstate__(self) -> State:
        dict_state, slots_state = super().__getstate__()
        dict_state['_individual'] = self._individual
        dict_state['_affiliation'] = self._affiliation
        return dict_state, slots_state

    def _get_privacy(self) -> Privacy:
        privacy = super()._get_privacy()
        if self.person:
            return merge_privacies(privacy, self.person.privacy)
        return privacy

    def __repr__(self) -> str:
        return repr_instance(self, id=self.id, individual=self.individual, affiliation=self.affiliation)

    def __eq__(self, other: Any) -> bool:
        if not super().__eq__(other):
            return False
        return (self._individual or '') == (other._individual or '') and (self._affiliation or '') == (other._affiliation or '')

    def __gt__(self, other: Any) -> bool:
        if other is None:
            return True
        if not isinstance(other, PersonName):
            return NotImplemented
        return (self._individual or '') > (other._individual or '') and (self._affiliation or '') > (other._affiliation or '')

    @classmethod
    def entity_type_label(cls) -> Str:
        return Str._('Person name')

    @classmethod
    def entity_type_label_plural(cls) -> Str:
        return Str._('Person names')

    @property
    def individual(self) -> str | None:
        return self._individual

    @property
    def affiliation(self) -> str | None:
        return self._affiliation

    @property
    def label(self) -> Str:
        if self.private:
            return Str._('private')
        return Str._(
            '{individual_name} {affiliation_name}',
            individual_name='…' if not self.individual else self.individual,
            affiliation_name='…' if not self.affiliation else self.affiliation,
        )


@total_ordering
@many_to_many('parents', 'betty.model.ancestry.Person', 'children')
@many_to_many('children', 'betty.model.ancestry.Person', 'parents')
@one_to_many('presences', 'betty.model.ancestry.Presence', 'person')
@one_to_many('names', 'betty.model.ancestry.PersonName', 'person')
class Person(HasFiles, HasCitations, HasLinks, HasMutablePrivacy, UserFacingEntity, Entity):
    def __init__(
        self,
        *,
        id: str | None = None,
        files: Iterable[File] | None = None,
        citations: Iterable[Citation] | None = None,
        links: set[Link] | None = None,
        privacy: Privacy | None = None,
        public: bool | None = None,
        private: bool | None = None,
        parents: Iterable[Person] | None = None,
        children: Iterable[Person] | None = None,
    ):
        super().__init__(
            id,
            files=files,
            citations=citations,
            links=links,
            privacy=privacy,
            public=public,
            private=private,
        )
        if children is not None:
            self.children = children  # type: ignore[assignment]
        if parents is not None:
            self.parents = parents  # type: ignore[assignment]

    @property
    def parents(self) -> EntityCollection[Person]:  # type: ignore[empty-body]
        pass

    @parents.setter
    def parents(self, parents: Iterable[Person]) -> None:
        pass

    @parents.deleter
    def parents(self) -> None:
        pass

    @property
    def children(self) -> EntityCollection[Person]:  # type: ignore[empty-body]
        pass

    @children.setter
    def children(self, children: Iterable[Person]) -> None:
        pass

    @children.deleter
    def children(self) -> None:
        pass

    @property
    def presences(self) -> EntityCollection[Presence]:  # type: ignore[empty-body]
        pass

    @presences.setter
    def presences(self, presences: Iterable[Presence]) -> None:
        pass

    @presences.deleter
    def presences(self) -> None:
        pass

    @property
    def names(self) -> EntityCollection[PersonName]:  # type: ignore[empty-body]
        pass

    @names.setter
    def names(self, names: Iterable[PersonName]) -> None:
        pass

    @names.deleter
    def names(self) -> None:
        pass

    @classmethod
    def entity_type_label(cls) -> Str:
        return Str._('Person')

    @classmethod
    def entity_type_label_plural(cls) -> Str:
        return Str._('People')

    def __eq__(self, other: Any) -> bool:
        return super().__eq__(other) and self.name == other.name

    def __gt__(self, other: Any) -> bool:
        if not isinstance(other, Person):
            return NotImplemented
        if self.name is None and other.name is None:
            return False
        return self.name > other.name  # type: ignore[operator]

    @property
    def name(self) -> PersonName | None:
        try:
            return next(filter(is_public, self.names))
        except StopIteration:
            try:
                return self.names[0]
            except IndexError:
                return None

    @property
    def alternative_names(self) -> list[PersonName]:
        return [
            name
            for name
            in self.names
            if is_public(name)
        ][1:]

    def _shortcut_event(self, event_type: type[EventType]) -> Presence | None:
        for presence in self.presences:
            if not isinstance(presence.role, Subject):
                continue
            if presence.event is None:
                continue
            if not issubclass(presence.event.event_type, event_type):
                continue
            if not presence.public:
                continue
            return presence
        return None

    @property
    def start(self) -> Presence | None:
        return self._shortcut_event(StartOfLifeEventType)

    @property
    def end(self) -> Presence | None:
        return self._shortcut_event(EndOfLifeEventType)

    @property
    def siblings(self) -> list[Person]:
        siblings = []
        for parent in self.parents:
            for sibling in parent.children:
                if sibling != self and sibling not in siblings:
                    siblings.append(sibling)
        return siblings

    @property
    def associated_files(self) -> Iterable[File]:
        files = [
            *self.files,
            *[
                file
                for name
                in self.names
                for citation
                in name.citations
                for file
                in citation.associated_files
            ],
            *[
                file
                for presence
                in self.presences
                if presence.event is not None
                for file
                in presence.event.associated_files
            ]
        ]
        # Preserve the original order.
        seen = set()
        for file in files:
            if file in seen:
                continue
            seen.add(file)
            yield file

    @property
    def label(self) -> Str:
        return self.name.label if self.name else super().label


class Ancestry(MultipleTypesEntityCollection[Entity]):
    def __init__(self):
        super().__init__()
        self._check_graph = True

    def __getstate__(self) -> PickleableEntityGraph:
        return PickleableEntityGraph(*self)

    def __setstate__(self, state: PickleableEntityGraph) -> None:
        self._collections = {}
        self.add_unchecked_graph(*state.build())

    def add_unchecked_graph(self, *entities: Entity) -> None:
        self._check_graph = False
        try:
            self.add(*entities)
        finally:
            self._check_graph = True

    def _on_add(self, *entities: Entity) -> None:
        super()._on_add(*entities)
        if self._check_graph:
            self.add(*self._get_associates(*entities))

    def _get_associates(self, *entities: Entity) -> Iterable[Entity]:
        for entity in entities:
            for association in EntityTypeAssociationRegistry.get_all_associations(entity):
                for associate in EntityTypeAssociationRegistry.get_associates(entity, association):
                    yield associate



from __future__ import annotations

import json
from typing import cast, Sequence

import yaml

from betty.locale import Str, Localizer
from betty.serde.dump import Dump, VoidableDump
from betty.serde.load import FormatError


class Format:
    @property
    def extensions(self) -> set[str]:
        raise NotImplementedError(repr(self))

    @property
    def label(self) -> Str:
        raise NotImplementedError(repr(self))

    def load(self, dump: str) -> Dump:
        raise NotImplementedError(repr(self))

    def dump(self, dump: VoidableDump) -> str:
        raise NotImplementedError(repr(self))


class Json(Format):
    @property
    def extensions(self) -> set[str]:
        return {'json'}

    @property
    def label(self) -> Str:
        return Str.plain('JSON')

    def load(self, dump: str) -> Dump:
        try:
            return cast(Dump, json.loads(dump))
        except json.JSONDecodeError as e:
            raise FormatError(Str._(
                'Invalid JSON: {error}.',
                error=str(e),
            ))

    def dump(self, dump: VoidableDump) -> str:
        return json.dumps(dump)


class Yaml(Format):
    @property
    def extensions(self) -> set[str]:
        return {'yaml', 'yml'}

    @property
    def label(self) -> Str:
        return Str.plain('YAML')

    def load(self, dump: str) -> Dump:
        try:
            return cast(Dump, yaml.safe_load(dump))
        except yaml.YAMLError as e:
            raise FormatError(Str._(
                'Invalid YAML: {error}.',
                error=str(e),
            ))

    def dump(self, dump: VoidableDump) -> str:
        return yaml.safe_dump(dump)


class FormatRepository:
    def __init__(
        self,
    ):
        super().__init__()
        self._formats = (
            Json(),
            Yaml(),
        )

    @property
    def formats(self) -> tuple[Format, ...]:
        return self._formats

    @property
    def extensions(self) -> tuple[str, ...]:
        return tuple(
            extension
            for _format in self._formats
            for extension in _format.extensions
        )

    def format_for(self, extension: str) -> Format:
        for format in self._formats:
            if extension in format.extensions:
                return format
        raise FormatError(Str._(
            'Unknown file format ".{extension}". Supported formats are: {supported_formats}.',
            extension=extension,
            supported_formats=FormatStr(self.formats),
        ))


class FormatStr(Str):
    def __init__(self, formats: Sequence[Format]):
        self._formats = formats

    def localize(self, localizer: Localizer) -> str:
        return ', '.join([
            f'.{extension} ({format.label.localize(localizer)})'
            for format in self._formats
            for extension in format.extensions
        ])



from __future__ import annotations

from contextlib import suppress
from pathlib import Path
from reprlib import recursive_repr
from typing import Any, Generic, final, Iterable, cast, Self
from urllib.parse import urlparse

from reactives import scope
from reactives.instance.property import reactive_property

from betty.app.extension import Extension, ConfigurableExtension
from betty.classtools import repr_instance
from betty.config import Configuration, Configurable, FileBasedConfiguration, ConfigurationMapping, \
    ConfigurationSequence
from betty.locale import get_data, Str
from betty.model import Entity, get_entity_type_name, UserFacingEntity, EntityT
from betty.model.ancestry import Ancestry, Person, Event, Place, Source
from betty.serde.dump import Dump, VoidableDump, void_none, minimize, Void, VoidableDictDump
from betty.serde.load import AssertionFailed, Fields, Assertions, Assertion, RequiredField, OptionalField, \
    Asserter

DEFAULT_LIFETIME_THRESHOLD = 125


class EntityReference(Configuration, Generic[EntityT]):
    def __init__(
        self,
        entity_type: type[EntityT] | None = None,
        entity_id: str | None = None,
        *,
        entity_type_is_constrained: bool = False,
    ):
        super().__init__()
        self._entity_type = entity_type
        self._entity_id = entity_id
        self._entity_type_is_constrained = entity_type_is_constrained

    @property
    @reactive_property
    def entity_type(self) -> type[EntityT] | None:
        return self._entity_type

    @entity_type.setter
    def entity_type(self, entity_type: type[EntityT]) -> None:
        if self._entity_type_is_constrained:
            raise AttributeError(f'The entity type cannot be set, as it is already constrained to {self._entity_type}.')
        self._entity_type = entity_type

    @property
    @reactive_property
    def entity_id(self) -> str | None:
        return self._entity_id

    @entity_id.setter
    def entity_id(self, entity_id: str) -> None:
        self._entity_id = entity_id

    @entity_id.deleter
    def entity_id(self) -> None:
        self._entity_id = None

    @property
    def entity_type_is_constrained(self) -> bool:
        return self._entity_type_is_constrained

    def update(self, other: Self) -> None:
        self._entity_type = other._entity_type
        self._entity_type_is_constrained = other._entity_type_is_constrained
        self._entity_id = other._entity_id
        self.react.trigger()

    @classmethod
    def load(
        cls,
        dump: Dump,
        configuration: Self | None = None,
    ) -> Self:
        if configuration is None:
            configuration = cls()
        asserter = Asserter()
        if isinstance(dump, dict) or not configuration.entity_type_is_constrained:
            asserter.assert_record(Fields(
                RequiredField(
                    'entity_type',
                    Assertions(asserter.assert_entity_type()) | asserter.assert_setattr(configuration, 'entity_type'),
                ),
                OptionalField(
                    'entity_id',
                    Assertions(asserter.assert_str()) | asserter.assert_setattr(configuration, 'entity_id'),
                ),
            ))(dump)
        else:
            asserter.assert_str()(dump)
            asserter.assert_setattr(configuration, 'entity_id')(dump)  # type: ignore[arg-type]
        return configuration

    def dump(self) -> VoidableDump:
        if self.entity_type_is_constrained:
            return void_none(self.entity_id)

        if self.entity_type is None or self.entity_id is None:
            return Void

        dump: VoidableDictDump[VoidableDump] = {
            'entity_type': get_entity_type_name(self._entity_type) if self._entity_type else Void,
            'entity_id': self._entity_id,
        }

        return minimize(dump)

    @scope.register_self
    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, EntityReference):
            return NotImplemented
        return self.entity_type == other.entity_type and self.entity_id == other.entity_id


class EntityReferenceSequence(Generic[EntityT], ConfigurationSequence[EntityReference[EntityT]]):
    def __init__(
        self,
        entity_references: Iterable[EntityReference[EntityT]] | None = None,
        *,
        entity_type_constraint: type[EntityT] | None = None,
    ):
        self._entity_type_constraint = entity_type_constraint
        super().__init__(entity_references)

    @classmethod
    def _item_type(cls) -> type[EntityReference[EntityT]]:
        return EntityReference

    def _on_add(self, configuration: EntityReference[EntityT]) -> None:
        super()._on_add(configuration)

        entity_type_constraint = self._entity_type_constraint
        entity_reference_entity_type = configuration._entity_type

        if entity_type_constraint is None:
            return

        if entity_reference_entity_type == entity_type_constraint and configuration.entity_type_is_constrained:
            return

        expected_entity_type_name = get_entity_type_name(
            cast(type[Entity], entity_type_constraint),
        )
        expected_entity_type_label = entity_type_constraint.entity_type_label()

        if entity_reference_entity_type is None:
            raise AssertionFailed(Str._(
                'The entity reference must be for an entity of type {expected_entity_type_name} ({expected_entity_type_label}), but instead does not specify an entity type at all.',
                expected_entity_type_name=expected_entity_type_name,
                expected_entity_type_label=expected_entity_type_label,
            ))

        actual_entity_type_label = entity_type_constraint.entity_type_label()

        raise AssertionFailed(Str._(
            'The entity reference must be for an entity of type {expected_entity_type_name} ({expected_entity_type_label}), but instead is for an entity of type {actual_entity_type_name} ({actual_entity_type_label})',
            expected_entity_type_name=expected_entity_type_name,
            expected_entity_type_label=expected_entity_type_label,
            actual_entity_type_name=get_entity_type_name(entity_reference_entity_type),
            actual_entity_type_label=actual_entity_type_label,
        ))


class ExtensionConfiguration(Configuration):
    def __init__(
        self,
        extension_type: type[Extension],
        enabled: bool = True,
        extension_configuration: Configuration | None = None,
    ):
        super().__init__()
        self._extension_type = extension_type
        self._enabled = enabled
        if extension_configuration is None and issubclass(extension_type, ConfigurableExtension):
            extension_configuration = extension_type.default_configuration()
        if extension_configuration is not None:
            extension_configuration.react(self)
        self._extension_configuration = extension_configuration

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        if self.extension_type != other.extension_type:
            return False
        if self.enabled != other.enabled:
            return False
        if self.extension_configuration != other.extension_configuration:
            return False
        return True

    @property
    def extension_type(self) -> type[Extension]:
        return self._extension_type

    @property
    @reactive_property
    def enabled(self) -> bool:
        return self._enabled

    @enabled.setter
    def enabled(self, enabled: bool) -> None:
        self._enabled = enabled

    @property
    def extension_configuration(self) -> Configuration | None:
        return self._extension_configuration

    def update(self, other: Self) -> None:
        raise NotImplementedError(repr(self))

    @classmethod
    def load(
        cls,
        dump: Dump,
        configuration: Self | None = None,
    ) -> Self:
        asserter = Asserter()
        extension_type = asserter.assert_field(RequiredField(
            'extension',
            Assertions(asserter.assert_extension_type()),
        ))(dump)
        if configuration is None:
            configuration = cls(extension_type)
        else:
            # This MUST NOT fail. If it does, this is a bug in the calling code that must be fixed.
            assert extension_type is configuration.extension_type
        asserter.assert_record(Fields(
            RequiredField(
                'extension',
            ),
            OptionalField(
                'enabled',
                Assertions(asserter.assert_bool()) | asserter.assert_setattr(configuration, 'enabled'),
            ),
            OptionalField(
                'configuration',
                Assertions(configuration._assert_load_extension_configuration(configuration.extension_type)),
            ),
        ))(dump)
        return configuration

    def _assert_load_extension_configuration(self, extension_type: type[Extension]) -> Assertion[Any, Configuration]:
        def _assertion(value: Any) -> Configuration:
            extension_configuration = self._extension_configuration
            if isinstance(extension_configuration, Configuration):
                return extension_configuration.load(value, extension_configuration)
            raise AssertionFailed(Str._(
                '{extension_type} is not configurable.',
                extension_type=extension_type.name(),
            ))
        return _assertion

    def dump(self) -> VoidableDump:
        return minimize({
            'extension': self.extension_type.name(),
            'enabled': self.enabled,
            'configuration': minimize(self.extension_configuration.dump()) if issubclass(self.extension_type, Configurable) and self.extension_configuration else Void,
        })


class ExtensionConfigurationMapping(ConfigurationMapping[type[Extension], ExtensionConfiguration]):
    def _minimize_item_dump(self) -> bool:
        return True

    @classmethod
    def _create_default_item(cls, configuration_key: type[Extension]) -> ExtensionConfiguration:
        return ExtensionConfiguration(configuration_key)

    def __init__(
        self,
        configurations: Iterable[ExtensionConfiguration] | None = None,
    ):
        super().__init__(configurations)

    @classmethod
    def _item_type(cls) -> type[ExtensionConfiguration]:
        return ExtensionConfiguration

    def _get_key(self, configuration: ExtensionConfiguration) -> type[Extension]:
        return configuration.extension_type

    @classmethod
    def _load_key(
        cls,
        item_dump: Dump,
        key_dump: str,
    ) -> Dump:
        asserter = Asserter()
        dict_dump = asserter.assert_dict()(item_dump)
        dict_dump['extension'] = key_dump
        return dict_dump

    def _dump_key(self, item_dump: VoidableDump) -> tuple[VoidableDump, str]:
        dict_dump = self._asserter.assert_dict()(item_dump)
        return dict_dump, dict_dump.pop('extension')

    def enable(self, *extension_types: type[Extension]) -> None:
        for extension_type in extension_types:
            try:
                self._configurations[extension_type].enabled = True
            except KeyError:
                self.append(ExtensionConfiguration(extension_type, True))

    def disable(self, *extension_types: type[Extension]) -> None:
        for extension_type in extension_types:
            with suppress(KeyError):
                self._configurations[extension_type].enabled = False


class EntityTypeConfiguration(Configuration):
    def __init__(self, entity_type: type[Entity], generate_html_list: bool | None = None):
        super().__init__()
        self._entity_type = entity_type
        self.generate_html_list = generate_html_list  # type: ignore[assignment]

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        if self.entity_type != other.entity_type:
            return False
        if self.generate_html_list != other.generate_html_list:
            return False
        return True

    @property
    def entity_type(self) -> type[Entity]:
        return self._entity_type

    @property
    @reactive_property
    def generate_html_list(self) -> bool:
        return self._generate_html_list or False

    @generate_html_list.setter
    def generate_html_list(self, generate_html_list: bool | None) -> None:
        if generate_html_list and not issubclass(self._entity_type, UserFacingEntity):
            raise AssertionFailed(Str._(
                'Cannot generate pages for {entity_type}, because it is not a user-facing entity type.',
                entity_type=get_entity_type_name(self._entity_type)
            ))
        self._generate_html_list = generate_html_list

    def update(self, other: Self) -> None:
        self._entity_type = other._entity_type
        self._generate_html_list = other._generate_html_list
        self.react.trigger()

    @classmethod
    def load(
            cls,
            dump: Dump,
            configuration: Self | None = None,
    ) -> Self:
        asserter = Asserter()
        entity_type = asserter.assert_field(RequiredField[Any, type[Entity]](
            'entity_type',
            Assertions(asserter.assert_str()) | asserter.assert_entity_type()),
        )(dump)
        configuration = cls(entity_type)
        asserter.assert_record(Fields(
            OptionalField(
                'entity_type',
            ),
            OptionalField(
                'generate_html_list',
                Assertions(asserter.assert_bool()) | asserter.assert_setattr(configuration, 'generate_html_list'),
            ),
        ))(dump)
        return configuration

    def dump(self) -> VoidableDump:
        dump: VoidableDictDump[VoidableDump] = {
            'entity_type': get_entity_type_name(self._entity_type),
            'generate_html_list': Void if self._generate_html_list is None else self._generate_html_list,
        }

        return minimize(dump)


class EntityTypeConfigurationMapping(ConfigurationMapping[type[Entity], EntityTypeConfiguration]):
    def _minimize_item_dump(self) -> bool:
        return True

    def _get_key(self, configuration: EntityTypeConfiguration) -> type[Entity]:
        return configuration.entity_type

    @classmethod
    def _load_key(
        cls,
        item_dump: Dump,
        key_dump: str,
    ) -> Dump:
        asserter = Asserter()
        dict_dump = asserter.assert_dict()(item_dump)
        asserter.assert_entity_type()(key_dump)
        dict_dump['entity_type'] = key_dump
        return dict_dump

    def _dump_key(self, item_dump: VoidableDump) -> tuple[VoidableDump, str]:
        dict_dump = self._asserter.assert_dict()(item_dump)
        return dict_dump, dict_dump.pop('entity_type')

    @classmethod
    def _item_type(cls) -> type[EntityTypeConfiguration]:
        return EntityTypeConfiguration

    @classmethod
    def _create_default_item(cls, configuration_key: type[Entity]) -> EntityTypeConfiguration:
        return EntityTypeConfiguration(configuration_key)


class LocaleConfiguration(Configuration):
    def __init__(self, locale: str, alias: str | None = None):
        super().__init__()
        self._locale = locale
        if alias is not None and '/' in alias:
            raise AssertionFailed(Str._('Locale aliases must not contain slashes.'))
        self._alias = alias

    @recursive_repr()
    def __repr__(self) -> str:
        return repr_instance(self, locale=self.locale, alias=self.alias)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        if self.locale != other.locale:
            return False
        if self.alias != other.alias:
            return False
        return True

    def __hash__(self) -> int:
        return hash((self._locale, self._alias))

    @property
    def locale(self) -> str:
        return self._locale

    @property
    @reactive_property
    def alias(self) -> str:
        if self._alias is None:
            return self.locale
        return self._alias

    @alias.setter
    def alias(self, alias: str | None) -> None:
        self._alias = alias

    def update(self, other: Self) -> None:
        self._locale = other._locale
        self._alias = other._alias

    @classmethod
    def load(
            cls,
            dump: Dump,
            configuration: Self | None = None,
    ) -> Self:
        asserter = Asserter()
        locale = asserter.assert_field(RequiredField(
            'locale',
            Assertions(asserter.assert_locale())),
        )(dump)
        if configuration is None:
            configuration = cls(locale)
        asserter.assert_record(Fields(
            RequiredField(
                'locale'
            ),
            OptionalField(
                'alias',
                Assertions(asserter.assert_str()) | asserter.assert_setattr(configuration, 'alias'),
            ),
        ))(dump)
        return configuration

    def dump(self) -> VoidableDump:
        return minimize({
            'locale': self.locale,
            'alias': void_none(self._alias)
        })


class LocaleConfigurationMapping(ConfigurationMapping[str, LocaleConfiguration]):
    @classmethod
    def _create_default_item(cls, configuration_key: str) -> LocaleConfiguration:
        return LocaleConfiguration(configuration_key)

    def __init__(
        self,
        configurations: Iterable[LocaleConfiguration] | None = None,
    ):
        super().__init__(configurations)
        if len(self) == 0:
            self.append(LocaleConfiguration('en-US'))

    def _get_key(self, configuration: LocaleConfiguration) -> str:
        return configuration.locale

    @classmethod
    def _load_key(
        cls,
        item_dump: Dump,
        key_dump: str,
    ) -> Dump:
        asserter = Asserter()
        dict_item_dump = asserter.assert_dict()(item_dump)
        dict_item_dump['locale'] = key_dump
        return dict_item_dump

    def _dump_key(self, item_dump: VoidableDump) -> tuple[VoidableDump, str]:
        dict_item_dump = self._asserter.assert_dict()(item_dump)
        return dict_item_dump, dict_item_dump.pop('locale')

    @classmethod
    def _item_type(cls) -> type[LocaleConfiguration]:
        return LocaleConfiguration

    def _on_remove(self, configuration: LocaleConfiguration) -> None:
        if len(self._configurations) <= 1:
            raise AssertionFailed(Str._(
                'Cannot remove the last remaining locale {locale}',
                locale=get_data(configuration.locale).get_display_name() or configuration.locale,
            ))

    @property
    @reactive_property
    def default(self) -> LocaleConfiguration:
        return next(iter(self._configurations.values()))

    @default.setter
    def default(self, configuration: LocaleConfiguration | str) -> None:
        if isinstance(configuration, str):
            configuration = self[configuration]
        self._configurations[configuration.locale] = configuration
        self._configurations.move_to_end(configuration.locale, False)
        self.react.trigger()

    @property
    def multilingual(self) -> bool:
        return len(self) > 1


@final
class ProjectConfiguration(FileBasedConfiguration):
    def __init__(self, base_url: str | None = None):
        super().__init__()
        self._base_url = 'https://example.com' if base_url is None else base_url
        self._root_path = ''
        self._clean_urls = False
        self._content_negotiation = False
        self._title = 'Betty'
        self._author: str | None = None
        self._entity_types = EntityTypeConfigurationMapping([
            EntityTypeConfiguration(Person, True),
            EntityTypeConfiguration(Event, True),
            EntityTypeConfiguration(Place, True),
            EntityTypeConfiguration(Source, True),
        ])
        self._entity_types.react(self)
        self._extensions = ExtensionConfigurationMapping()
        self._extensions.react(self)
        self._debug = False
        self._locales = LocaleConfigurationMapping()
        self._locales.react(self)
        self._lifetime_threshold = DEFAULT_LIFETIME_THRESHOLD

    @property
    def project_directory_path(self) -> Path:
        return self.configuration_file_path.parent

    @property
    def output_directory_path(self) -> Path:
        return self.project_directory_path / 'output'

    @property
    def assets_directory_path(self) -> Path:
        return self.project_directory_path / 'assets'

    @property
    def www_directory_path(self) -> Path:
        return self.output_directory_path / 'www'

    def localize_www_directory_path(self, locale: str) -> Path:
        if self.locales.multilingual:
            return self.www_directory_path / self.locales[locale].alias
        return self.www_directory_path

    @property
    @reactive_property
    def title(self) -> str:
        return self._title

    @title.setter
    def title(self, title: str) -> None:
        self._title = title

    @property
    @reactive_property
    def author(self) -> str | None:
        return self._author

    @author.setter
    def author(self, author: str | None) -> None:
        self._author = author

    @property
    @reactive_property
    def base_url(self) -> str:
        return self._base_url

    @base_url.setter
    def base_url(self, base_url: str) -> None:
        base_url_parts = urlparse(base_url)
        if not base_url_parts.scheme:
            raise AssertionFailed(Str._('The base URL must start with a scheme such as https://, http://, or file://.'))
        if not base_url_parts.netloc:
            raise AssertionFailed(Str._('The base URL must include a path.'))
        self._base_url = '%s://%s' % (base_url_parts.scheme, base_url_parts.netloc)

    @property
    @reactive_property
    def root_path(self) -> str:
        return self._root_path

    @root_path.setter
    def root_path(self, root_path: str) -> None:
        self._root_path = root_path.strip('/')

    @property
    @reactive_property
    def content_negotiation(self) -> bool:
        return self._content_negotiation

    @content_negotiation.setter
    def content_negotiation(self, content_negotiation: bool) -> None:
        self._content_negotiation = content_negotiation

    @property
    @reactive_property
    def clean_urls(self) -> bool:
        return self._clean_urls or self.content_negotiation

    @clean_urls.setter
    def clean_urls(self, clean_urls: bool) -> None:
        self._clean_urls = clean_urls

    @property
    def locales(self) -> LocaleConfigurationMapping:
        return self._locales

    @property
    def entity_types(self) -> EntityTypeConfigurationMapping:
        return self._entity_types

    @property
    def extensions(self) -> ExtensionConfigurationMapping:
        return self._extensions

    @property
    @reactive_property
    def debug(self) -> bool:
        return self._debug

    @debug.setter
    def debug(self, debug: bool) -> None:
        self._debug = debug

    @property
    @reactive_property
    def lifetime_threshold(self) -> int:
        return self._lifetime_threshold

    @lifetime_threshold.setter
    def lifetime_threshold(self, lifetime_threshold: int) -> None:
        self._asserter.assert_positive_number()(lifetime_threshold)
        self._lifetime_threshold = lifetime_threshold

    def update(self, other: Self) -> None:
        self._base_url = other._base_url
        self._title = other._title
        self._author = other._author
        self._root_path = other._root_path
        self._clean_urls = other._clean_urls
        self._content_negotiation = other._content_negotiation
        self._debug = other._debug
        self._lifetime_threshold = other._lifetime_threshold
        self._locales.update(other._locales)
        self._extensions.update(other._extensions)
        self._entity_types.update(other._entity_types)

    @classmethod
    def load(
            cls,
            dump: Dump,
            configuration: Self | None = None,
    ) -> Self:
        if configuration is None:
            configuration = cls()
        asserter = Asserter()
        asserter.assert_record(Fields(
            RequiredField(
                'base_url',
                Assertions(asserter.assert_str()) | asserter.assert_setattr(configuration, 'base_url'),
            ),
            OptionalField(
                'title',
                Assertions(asserter.assert_str()) | asserter.assert_setattr(configuration, 'title'),
            ),
            OptionalField(
                'author',
                Assertions(asserter.assert_str()) | asserter.assert_setattr(configuration, 'author'),
            ),
            OptionalField(
                'root_path',
                Assertions(asserter.assert_str()) | asserter.assert_setattr(configuration, 'root_path'),
            ),
            OptionalField(
                'clean_urls',
                Assertions(asserter.assert_bool()) | asserter.assert_setattr(configuration, 'clean_urls'),
            ),
            OptionalField(
                'content_negotiation',
                Assertions(asserter.assert_bool()) | asserter.assert_setattr(configuration, 'content_negotiation'),
            ),
            OptionalField(
                'debug',
                Assertions(asserter.assert_bool()) | asserter.assert_setattr(configuration, 'debug'),
            ),
            OptionalField(
                'lifetime_threshold',
                Assertions(asserter.assert_int()) | asserter.assert_setattr(configuration, 'lifetime_threshold'),
            ),
            OptionalField(
                'locales',
                Assertions(configuration._locales.assert_load(configuration.locales)),
            ),
            OptionalField(
                'extensions',
                Assertions(configuration._extensions.assert_load(configuration.extensions)),
            ),
            OptionalField(
                'entity_types',
                Assertions(configuration._entity_types.assert_load(configuration.entity_types)),
            ),
        ))(dump)
        return configuration

    def dump(self) -> VoidableDictDump[Dump]:
        return minimize({  # type: ignore[return-value]
            'base_url': self.base_url,
            'title': self.title,
            'root_path': void_none(self.root_path),
            'clean_urls': void_none(self.clean_urls),
            'author': void_none(self.author),
            'content_negotiation': void_none(self.content_negotiation),
            'debug': void_none(self.debug),
            'lifetime_threshold': void_none(self.lifetime_threshold),
            'locales': self.locales.dump(),
            'extensions': self.extensions.dump(),
            'entity_types': self.entity_types.dump(),
        }, True)


class Project(Configurable[ProjectConfiguration]):
    def __init__(self):
        super().__init__()
        self._configuration = ProjectConfiguration()
        self._ancestry = Ancestry()

    def __getstate__(self) -> tuple[VoidableDump, Path, Ancestry]:
        return self._configuration.dump(), self._configuration.configuration_file_path, self._ancestry

    def __setstate__(self, state: tuple[Dump, Path, Ancestry]) -> None:
        dump, configuration_file_path, self._ancestry = state
        self._configuration = ProjectConfiguration.load(dump)
        self._configuration.configuration_file_path = configuration_file_path

    @property
    def ancestry(self) -> Ancestry:
        return self._ancestry



from __future__ import annotations

import gzip
import re
import tarfile
from collections import defaultdict
from contextlib import suppress
from pathlib import Path
from typing import Iterable, Any, IO, cast
from xml.etree import ElementTree

import aiofiles
from aiofiles.tempfile import TemporaryDirectory
from geopy import Point

from betty.gramps.error import GrampsError
from betty.load import getLogger
from betty.locale import DateRange, Datey, Date, Str, Localizer
from betty.media_type import MediaType
from betty.model import Entity, EntityGraphBuilder, AliasedEntity, AliasableEntity
from betty.model.ancestry import Ancestry, Note, File, Source, Citation, Place, Event, Person, PersonName, Subject, \
    Witness, Beneficiary, Attendee, Presence, PlaceName, Enclosure, HasLinks, Link, HasFiles, HasCitations, \
    HasMutablePrivacy, Speaker, Celebrant, Organizer
from betty.model.event_type import Birth, Baptism, Adoption, Cremation, Death, Funeral, Burial, Will, Engagement, \
    Marriage, MarriageAnnouncement, Divorce, DivorceAnnouncement, Residence, Immigration, Emigration, Occupation, \
    Retirement, Correspondence, Confirmation, Missing, UnknownEventType, EventType, Conference
from betty.path import rootname


class GrampsLoadFileError(GrampsError, RuntimeError):
    pass


class GrampsFileNotFoundError(GrampsError, FileNotFoundError):
    pass


class XPathError(GrampsError, RuntimeError):
    pass


class GrampsLoader:
    def __init__(
        self,
        ancestry: Ancestry,
            *,
            localizer: Localizer,
    ):
        super().__init__()
        self._ancestry = ancestry
        self._ancestry_builder = EntityGraphBuilder()
        self._added_entity_counts: dict[type[Entity], int] = defaultdict(lambda: 0)
        self._tree: ElementTree.ElementTree | None = None
        self._gramps_tree_directory_path: Path | None = None
        self._loaded = False
        self._localizer = localizer

    async def load_file(self, file_path: Path) -> None:
        file_path = file_path.resolve()
        logger = getLogger()
        logger.info(self._localizer._('Loading "{file_path}"...').format(
            file_path=str(file_path),
        ))

        with suppress(GrampsLoadFileError):
            await self.load_gpkg(file_path)
            return

        with suppress(GrampsLoadFileError):
            await self.load_gramps(file_path)
            return

        try:
            async with aiofiles.open(file_path) as f:
                xml = await f.read()
        except FileNotFoundError:
            raise GrampsFileNotFoundError(Str._(
                'Could not find the file "{file_path}".',
                file_path=str(file_path),
            )) from None
        with suppress(GrampsLoadFileError):
            await self.load_xml(xml, Path(file_path.anchor))
            return

        raise GrampsLoadFileError(Str._(
            'Could not load "{file_path}" as a *.gpkg, a *.gramps, or an *.xml family tree.',
            file_path=str(file_path),
        ))

    async def load_gramps(self, gramps_path: Path) -> None:
        gramps_path = gramps_path.resolve()
        try:
            with gzip.open(gramps_path, mode='r') as f:
                xml: str = f.read()  # type: ignore[assignment]
            await self.load_xml(
                xml,
                rootname(gramps_path),
            )
        except OSError as e:
            raise GrampsLoadFileError(Str.plain(e))

    async def load_gpkg(self, gpkg_path: Path) -> None:
        gpkg_path = gpkg_path.resolve()
        try:
            tar_file: IO[bytes] = gzip.open(gpkg_path)  # type: ignore[assignment]
            try:
                async with TemporaryDirectory() as cache_directory_path_str:
                    tarfile.open(
                        fileobj=tar_file,
                    ).extractall(cache_directory_path_str)
                    await self.load_gramps(Path(cache_directory_path_str) / 'data.gramps')
            except tarfile.ReadError:
                raise GrampsLoadFileError(Str._(
                    'Could not extract {file_path} as a tar (*.tar) file after extracting the outer gzip (*.gz) file.',
                    file_path=str(gpkg_path),
                ))
        except OSError:
            raise GrampsLoadFileError(Str._(
                'Could not extract {file_path} as a gzip (*.gz) file.',
                file_path=str(gpkg_path),
            ))

    async def load_xml(self, xml: str | Path, gramps_tree_directory_path: Path) -> None:
        if isinstance(xml, Path):
            async with aiofiles.open(xml) as f:
                xml = await f.read()
        try:
            tree = ElementTree.ElementTree(ElementTree.fromstring(
                xml,
            ))
        except ElementTree.ParseError as e:
            raise GrampsLoadFileError(Str.plain(e))
        await self.load_tree(tree, gramps_tree_directory_path)

    async def load_tree(self, tree: ElementTree.ElementTree, gramps_tree_directory_path: Path) -> None:
        if self._loaded:
            raise RuntimeError('This loader has been used up.')

        self._loaded = True
        self._tree = tree
        self._gramps_tree_directory_path = gramps_tree_directory_path.resolve()

        logger = getLogger()

        database = self._tree.getroot()

        self._load_notes(database)
        # @todo Localize all log messages
        logger.info(f'Loaded {self._added_entity_counts[Note]} notes.')

        self._load_objects(database, self._gramps_tree_directory_path)
        logger.info(f'Loaded {self._added_entity_counts[File]} files.')

        self._load_repositories(database)
        repository_count = self._added_entity_counts[Source]
        logger.info(f'Loaded {repository_count} repositories as sources.')

        self._load_sources(database)
        logger.info(f'Loaded {self._added_entity_counts[Source] - repository_count} sources.')

        self._load_citations(database)
        logger.info(f'Loaded {self._added_entity_counts[Citation]} citations.')

        self._load_places(database)
        logger.info(f'Loaded {self._added_entity_counts[Place]} places.')

        self._load_events(database)
        logger.info(f'Loaded {self._added_entity_counts[Event]} events.')

        self._load_people(database)
        logger.info(f'Loaded {self._added_entity_counts[Person]} people.')

        self._load_families(database)

        self._ancestry.add_unchecked_graph(*self._ancestry_builder.build())

    def add_entity(self, entity: AliasableEntity[Entity]) -> None:
        self._ancestry_builder.add_entity(entity)
        self._added_entity_counts[entity.type] += 1

    def add_association(self, *args: Any, **kwargs: Any) -> None:
        self._ancestry_builder.add_association(*args, **kwargs)

    _NS = {
        'ns': 'http://gramps-project.org/xml/1.7.1/',
    }

    def _xpath(self, element: ElementTree.Element, selector: str) -> list[ElementTree.Element]:
        return element.findall(selector, namespaces=self._NS)

    def _xpath1(self, element: ElementTree.Element, selector: str) -> ElementTree.Element:
        found_element = element.find(selector, namespaces=self._NS)
        if found_element is None:
            raise XPathError(Str.plain(
                'Cannot find an element "{selector}" within {element}.',
                selector=selector,
                element=str(element),
            ))
        return found_element

    _DATE_PATTERN = re.compile(r'^.{4}((-.{2})?-.{2})?$')
    _DATE_PART_PATTERN = re.compile(r'^\d+$')

    def _load_date(self, element: ElementTree.Element) -> Datey | None:
        with suppress(XPathError):
            dateval_element = self._xpath1(element, './ns:dateval')
            if dateval_element.get('cformat') is None:
                dateval_type = dateval_element.get('type')
                if dateval_type is None:
                    return self._load_dateval(dateval_element, 'val')
                dateval_type = str(dateval_type)
                if dateval_type == 'about':
                    date = self._load_dateval(dateval_element, 'val')
                    if date is None:
                        return None
                    date.fuzzy = True
                    return date
                if dateval_type == 'before':
                    return DateRange(
                        None,
                        self._load_dateval(dateval_element, 'val'),
                        end_is_boundary=True,
                    )
                if dateval_type == 'after':
                    return DateRange(
                        self._load_dateval(dateval_element, 'val'),
                        start_is_boundary=True,
                    )
        with suppress(XPathError):
            datespan_element = self._xpath1(element, './ns:datespan')
            if datespan_element.get('cformat') is None:
                return DateRange(
                    self._load_dateval(datespan_element, 'start'),
                    self._load_dateval(datespan_element, 'stop'),
                )
        with suppress(XPathError):
            daterange_element = self._xpath1(element, './ns:daterange')
            if daterange_element.get('cformat') is None:
                return DateRange(
                    self._load_dateval(daterange_element, 'start'),
                    self._load_dateval(daterange_element, 'stop'),
                    start_is_boundary=True,
                    end_is_boundary=True,
                )
        return None

    def _load_dateval(self, element: ElementTree.Element, value_attribute_name: str) -> Date | None:
        dateval = str(element.get(value_attribute_name))
        if self._DATE_PATTERN.fullmatch(dateval):
            date_parts: tuple[int | None, int | None, int | None] = tuple(  # type: ignore[assignment]
                int(part)
                if self._DATE_PART_PATTERN.fullmatch(part) and int(part) > 0
                else None
                for part
                in dateval.split('-')
            )
            date = Date(*date_parts)
            dateval_quality = element.get('quality')
            if dateval_quality == 'estimated':
                date.fuzzy = True
            return date
        return None

    def _load_notes(self, database: ElementTree.Element) -> None:
        for element in self._xpath(database, './ns:notes/ns:note'):
            self._load_note(element)

    def _load_note(self, element: ElementTree.Element) -> None:
        handle = element.get('handle')
        note_id = element.get('id')
        assert note_id is not None
        text_element = self._xpath1(element, './ns:text')
        assert text_element is not None
        text = str(text_element.text)
        self.add_entity(AliasedEntity(Note(
            id=note_id,
            text=text,
        ), handle))

    def _load_objects(self, database: ElementTree.Element, gramps_tree_directory_path: Path) -> None:
        for element in self._xpath(database, './ns:objects/ns:object'):
            self._load_object(element, gramps_tree_directory_path)

    def _load_object(self, element: ElementTree.Element, gramps_tree_directory_path: Path) -> None:
        file_handle = element.get('handle')
        file_id = element.get('id')
        file_element = self._xpath1(element, './ns:file')
        src = file_element.get('src')
        assert src is not None
        file_path = gramps_tree_directory_path / src
        file = File(
            id=file_id,
            path=file_path,
        )
        mime = file_element.get('mime')
        assert mime is not None
        file.media_type = MediaType(mime)
        description = file_element.get('description')
        if description:
            file.description = description
        self._load_attribute_privacy(file, element, 'attribute')
        self.add_entity(AliasedEntity(file, file_handle))
        for citation_handle in self._load_handles('citationref', element):
            self.add_association(File, file_handle, 'citations', Citation, citation_handle)
        for note_handle in self._load_handles('noteref', element):
            self.add_association(File, file_handle, 'notes', Note, note_handle)

    def _load_people(self, database: ElementTree.Element) -> None:
        for element in self._xpath(database, './ns:people/ns:person'):
            self._load_person(element)

    def _load_person(self, element: ElementTree.Element) -> None:
        person_handle = element.get('handle')
        assert person_handle is not None
        person = Person(id=element.get('id'))

        name_elements = sorted(self._xpath(element, './ns:name'), key=lambda x: x.get('alt') == '1')
        person_names = []
        for name_element in name_elements:
            is_alternative = name_element.get('alt') == '1'
            try:
                individual_name = self._xpath1(name_element, './ns:first').text
            except XPathError:
                individual_name = None
            surname_elements = [
                surname_element
                for surname_element
                in self._xpath(
                    name_element,
                    './ns:surname'
                )
                if surname_element.text is not None
            ]
            if surname_elements:
                for surname_element in surname_elements:
                    if not is_alternative:
                        is_alternative = surname_element.get('prim') == '0'
                    affiliation_name = surname_element.text
                    surname_prefix = surname_element.get('prefix')
                    if surname_prefix is not None:
                        affiliation_name = '%s %s' % (
                            surname_prefix, affiliation_name)
                    person_name = PersonName(
                        individual=individual_name,
                        affiliation=affiliation_name,
                    )
                    self._load_citationref(person_name, name_element)
                    person_names.append((person_name, is_alternative))
            elif individual_name is not None:
                person_name = PersonName(individual=individual_name)
                self._load_citationref(person_name, name_element)
                person_names.append((person_name, is_alternative))
        for person_name, _ in sorted(person_names, key=lambda x: x[1]):
            self.add_entity(person_name)
            self.add_association(Person, person_handle, 'names', PersonName, person_name.id)

        self._load_eventrefs(person_handle, element)
        if element.get('priv') == '1':
            person.private = True

        aliased_person = AliasedEntity(person, person_handle)
        self._load_citationref(
            aliased_person,  # type: ignore[arg-type]
            element,
        )
        self._load_objref(
            aliased_person,  # type: ignore[arg-type]
            element,
        )
        self._load_urls(person, element)
        self._load_attribute_privacy(person, element, 'attribute')
        self.add_entity(
            aliased_person,  # type: ignore[arg-type]
        )

    def _load_families(self, database: ElementTree.Element) -> None:
        for element in self._xpath(database, './ns:families/ns:family'):
            self._load_family(element)

    def _load_family(self, element: ElementTree.Element) -> None:
        parent_handles = []

        # Load the father.
        father_handle = self._load_handle('father', element)
        if father_handle is not None:
            self._load_eventrefs(father_handle, element)
            parent_handles.append(father_handle)

        # Load the mother.
        mother_handle = self._load_handle('mother', element)
        if mother_handle is not None:
            self._load_eventrefs(mother_handle, element)
            parent_handles.append(mother_handle)

        # Load the children.
        child_handles = self._load_handles('childref', element)
        for child_handle in child_handles:
            for parent_handle in parent_handles:
                self.add_association(Person, child_handle, 'parents', Person, parent_handle)

    def _load_eventrefs(self, person_id: str, element: ElementTree.Element) -> None:
        eventrefs = self._xpath(element, './ns:eventref')
        for eventref in eventrefs:
            self._load_eventref(person_id, eventref)

    _PRESENCE_ROLE_MAP = {
        'Primary': Subject(),
        'Family': Subject(),
        'Witness': Witness(),
        'Beneficiary': Beneficiary(),
        'Speaker': Speaker(),
        'Celebrant': Celebrant(),
        'Organizer': Organizer(),
        'Attendee': Attendee(),
        'Unknown': Attendee(),
    }

    def _load_eventref(self, person_id: str, eventref: ElementTree.Element) -> None:
        event_handle = eventref.get('hlink')
        assert event_handle is not None
        gramps_presence_role = cast(str, eventref.get('role'))

        try:
            role = self._PRESENCE_ROLE_MAP[gramps_presence_role]
        except KeyError:
            role = Attendee()
            getLogger().warning(
                Str._(
                    'Betty is unfamiliar with person "{person_id}"\'s Gramps presence role of "{gramps_presence_role}" for the event with Gramps handle "{event_handle}". The role was imported, but set to "{betty_presence_role}".',
                    person_id=person_id,
                    event_handle=event_handle,
                    gramps_presence_role=gramps_presence_role,
                    betty_presence_role=role.label,
                )
            )

        presence = Presence(None, role, None)
        self.add_entity(presence)
        self.add_association(Presence, presence.id, 'person', Person, person_id)
        self.add_association(Presence, presence.id, 'event', Event, event_handle)

    def _load_places(self, database: ElementTree.Element) -> None:
        for element in self._xpath(database, './ns:places/ns:placeobj'):
            self._load_place(element)

    def _load_place(self, element: ElementTree.Element) -> None:
        place_handle = element.get('handle')
        names = []
        for name_element in self._xpath(element, './ns:pname'):
            # The Gramps language is a single ISO language code, which is a valid BCP 47 locale.
            language = name_element.get('lang')
            date = self._load_date(name_element)
            name = name_element.get('value')
            assert name is not None
            names.append(PlaceName(
                name=name,
                locale=language,
                date=date,
            ))

        place = Place(
            id=element.get('id'),
            names=names,
        )

        coordinates = self._load_coordinates(element)
        if coordinates:
            place.coordinates = coordinates

        self._load_urls(place, element)

        self.add_entity(AliasedEntity(place, place_handle))

        for enclosed_by_handle in self._load_handles('placeref', element):
            aliased_enclosure = AliasedEntity(Enclosure(encloses=None, enclosed_by=None))
            self.add_entity(
                aliased_enclosure,  # type: ignore[arg-type]
            )
            self.add_association(Enclosure, aliased_enclosure.id, 'encloses', Place, place_handle)
            self.add_association(Enclosure, aliased_enclosure.id, 'enclosed_by', Place, enclosed_by_handle)

    def _load_coordinates(self, element: ElementTree.Element) -> Point | None:
        with suppress(XPathError):
            coord_element = self._xpath1(element, './ns:coord')

            coordinates = f'{coord_element.get("lat")}; {coord_element.get("long")}'
            try:
                return Point.from_string(coordinates)
            except ValueError:
                getLogger().warning(Str._(
                    'Cannot load coordinates "{coordinates}", because they are in an unknown format.',
                    coordinates=coordinates,
                ))
        return None

    def _load_events(self, database: ElementTree.Element) -> None:
        for element in self._xpath(database, './ns:events/ns:event'):
            self._load_event(element)

    _EVENT_TYPE_MAP = {
        'Birth': Birth,
        'Baptism': Baptism,
        'Adopted': Adoption,
        'Cremation': Cremation,
        'Death': Death,
        'Funeral': Funeral,
        'Burial': Burial,
        'Will': Will,
        'Engagement': Engagement,
        'Marriage': Marriage,
        'Marriage Banns': MarriageAnnouncement,
        'Divorce': Divorce,
        'Divorce Filing': DivorceAnnouncement,
        'Residence': Residence,
        'Immigration': Immigration,
        'Emigration': Emigration,
        'Occupation': Occupation,
        'Retirement': Retirement,
        'Correspondence': Correspondence,
        'Confirmation': Confirmation,
        'Missing': Missing,
        'Conference': Conference,
    }

    def _load_event(self, element: ElementTree.Element) -> None:
        event_handle = element.get('handle')
        event_id = element.get('id')
        assert event_id is not None
        gramps_type = self._xpath1(element, './ns:type').text
        assert gramps_type is not None

        try:
            event_type: type[EventType] = self._EVENT_TYPE_MAP[gramps_type]
        except KeyError:
            event_type = UnknownEventType
            getLogger().warning(
                Str._(
                    'Betty is unfamiliar with Gramps event "{event_id}"\'s type of "{gramps_event_type}". The event was imported, but its type was set to "{betty_event_type}".',
                    event_id=event_id,
                    gramps_event_type=gramps_type,
                    betty_event_type=event_type.label(),
                )
            )

        event = Event(
            id=event_id,
            event_type=event_type,
        )

        event.date = self._load_date(element)

        # Load the event place.
        place_handle = self._load_handle('place', element)
        if place_handle is not None:
            self.add_association(Event, event_handle, 'place', Place, place_handle)

        # Load the description.
        with suppress(XPathError):
            event.description = self._xpath1(element, './ns:description').text

        self._load_attribute_privacy(event, element, 'attribute')

        aliased_event = AliasedEntity(event, event_handle)
        self._load_objref(
            aliased_event,  # type: ignore[arg-type]
            element,
        )
        self._load_citationref(
            aliased_event,  # type: ignore[arg-type]
            element,
        )
        self.add_entity(
            aliased_event,  # type: ignore[arg-type]
        )

    def _load_repositories(self, database: ElementTree.Element) -> None:
        for element in self._xpath(database, './ns:repositories/ns:repository'):
            self._load_repository(element)

    def _load_repository(self, element: ElementTree.Element) -> None:
        repository_source_handle = element.get('handle')

        source = Source(
            id=element.get('id'),
            name=self._xpath1(element, './ns:rname').text,
        )

        self._load_urls(source, element)

        self.add_entity(AliasedEntity(source, repository_source_handle))

    def _load_sources(self, database: ElementTree.Element) -> None:
        for element in self._xpath(database, './ns:sources/ns:source'):
            self._load_source(element)

    def _load_source(self, element: ElementTree.Element) -> None:
        source_handle = element.get('handle')
        try:
            source_name = self._xpath1(element, './ns:stitle').text
        except XPathError:
            source_name = None

        source = Source(
            id=element.get('id'),
            name=source_name,
        )

        repository_source_handle = self._load_handle('reporef', element)
        if repository_source_handle is not None:
            self.add_association(Source, source_handle, 'contained_by', Source, repository_source_handle)

        # Load the author.
        with suppress(XPathError):
            source.author = self._xpath1(element, './ns:sauthor').text

        # Load the publication info.
        with suppress(XPathError):
            source.publisher = self._xpath1(element, './ns:spubinfo').text

        self._load_attribute_privacy(source, element, 'srcattribute')

        aliased_source = AliasedEntity(source, source_handle)
        self._load_objref(
            aliased_source,  # type: ignore[arg-type]
            element,
        )
        self.add_entity(
            aliased_source,  # type: ignore[arg-type]
        )

    def _load_citations(self, database: ElementTree.Element) -> None:
        for element in self._xpath(database, './ns:citations/ns:citation'):
            self._load_citation(element)

    def _load_citation(self, element: ElementTree.Element) -> None:
        citation_handle = element.get('handle')
        source_handle = self._xpath1(element, './ns:sourceref').get('hlink')

        citation = Citation(id=element.get('id'))
        self.add_association(Citation, citation_handle, 'source', Source, source_handle)

        citation.date = self._load_date(element)
        self._load_attribute_privacy(citation, element, 'srcattribute')

        with suppress(XPathError):
            citation.location = Str.plain(self._xpath1(element, './ns:page').text)

        aliased_citation = AliasedEntity(citation, citation_handle)
        self._load_objref(
            aliased_citation,  # type: ignore[arg-type]
            element,
        )
        self.add_entity(
            aliased_citation,  # type: ignore[arg-type]
        )

    def _load_citationref(self, owner: AliasableEntity[HasCitations & Entity], element: ElementTree.Element) -> None:
        for citation_handle in self._load_handles('citationref', element):
            self.add_association(owner.type, owner.id, 'citations', Citation, citation_handle)

    def _load_handles(self, handle_type: str, element: ElementTree.Element) -> Iterable[str]:
        for citation_handle_element in self._xpath(element, f'./ns:{handle_type}'):
            hlink = citation_handle_element.get('hlink')
            if hlink:
                yield hlink

    def _load_handle(self, handle_type: str, element: ElementTree.Element) -> str | None:
        for citation_handle_element in self._xpath(element, f'./ns:{handle_type}'):
            return citation_handle_element.get('hlink')
        return None

    def _load_objref(self, owner: AliasableEntity[HasFiles & Entity], element: ElementTree.Element) -> None:
        file_handles = self._load_handles('objref', element)
        for file_handle in file_handles:
            self.add_association(owner.type, owner.id, 'files', File, file_handle)

    def _load_urls(self, owner: HasLinks, element: ElementTree.Element) -> None:
        url_elements = self._xpath(element, './ns:url')
        for url_element in url_elements:
            link = Link(str(url_element.get('href')))
            link.relationship = 'external'
            link.label = url_element.get('description')
            owner.links.add(link)

    def _load_attribute_privacy(self, entity: HasMutablePrivacy & Entity, element: ElementTree.Element, tag: str) -> None:
        privacy_value = self._load_attribute('privacy', element, tag)
        if privacy_value is None:
            return
        if privacy_value == 'private':
            entity.private = True
            return
        if privacy_value == 'public':
            entity.public = True
            return
        getLogger().warning(
            Str._(
                'The betty:privacy Gramps attribute must have a value of "public" or "private", but "{privacy_value}" was given for {entity_type} {entity_id} ({entity_label}), which was ignored.',
                privacy_value=privacy_value,
                entity_type=entity.entity_type_label(),
                entity_id=entity.id,
                entity_label=entity.label,
            ))

    def _load_attribute(self, name: str, element: ElementTree.Element, tag: str) -> str | None:
        with suppress(XPathError):
            return self._xpath1(element, './ns:%s[@type="betty:%s"]' % (tag, name)).get('value')
        return None



import webbrowser
from datetime import datetime
from os import path
from pathlib import Path
from typing import Any
from urllib.parse import urlencode

from PyQt6.QtCore import Qt, QCoreApplication
from PyQt6.QtGui import QIcon, QAction
from PyQt6.QtWidgets import QFormLayout, QWidget, QVBoxLayout, QHBoxLayout, QFileDialog, QPushButton

from betty import about
from betty.about import report
from betty.asyncio import sync, wait
from betty.gui import BettyWindow, get_configuration_file_filter
from betty.gui.error import catch_exceptions
from betty.gui.locale import TranslationsLocaleCollector
from betty.gui.serve import ServeDemoWindow
from betty.gui.text import Text
from betty.project import ProjectConfiguration


class BettyMainWindow(BettyWindow):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.setWindowIcon(QIcon(path.join(path.dirname(__file__), 'assets', 'public', 'static', 'betty-512x512.png')))

        menu_bar = self.menuBar()
        assert menu_bar is not None

        betty_menu = menu_bar.addMenu('&Betty')
        assert betty_menu is not None
        self.betty_menu = betty_menu

        self.new_project_action = QAction(self)
        self.new_project_action.setShortcut('Ctrl+N')
        self.new_project_action.triggered.connect(
            lambda _: self.new_project(),
        )
        betty_menu.addAction(self.new_project_action)

        self.open_project_action = QAction(self)
        self.open_project_action.setShortcut('Ctrl+O')
        self.open_project_action.triggered.connect(
            lambda _: self.open_project(),
        )
        betty_menu.addAction(self.open_project_action)

        self._demo_action = QAction(self)
        self._demo_action.triggered.connect(
            lambda _: self._demo(),
        )
        betty_menu.addAction(self._demo_action)

        self.open_application_configuration_action = QAction(self)
        self.open_application_configuration_action.triggered.connect(
            lambda _: self.open_application_configuration(),
        )
        betty_menu.addAction(self.open_application_configuration_action)

        self.clear_caches_action = QAction(self)
        self.clear_caches_action.triggered.connect(
            lambda _: self.clear_caches(),
        )
        betty_menu.addAction(self.clear_caches_action)

        self.exit_action = QAction(self)
        self.exit_action.setShortcut('Ctrl+Q')
        self.exit_action.triggered.connect(QCoreApplication.quit)
        betty_menu.addAction(self.exit_action)

        help_menu = menu_bar.addMenu('')
        assert help_menu is not None
        self.help_menu = help_menu

        self.report_bug_action = QAction(self)
        self.report_bug_action.triggered.connect(
            lambda _: self.report_bug(),
        )
        help_menu.addAction(self.report_bug_action)

        self.request_feature_action = QAction(self)
        self.request_feature_action.triggered.connect(
            lambda _: self.request_feature(),
        )
        help_menu.addAction(self.request_feature_action)

        self.about_action = QAction(self)
        self.about_action.triggered.connect(
            lambda _: self._about_betty(),
        )
        help_menu.addAction(self.about_action)

    @property
    def title(self) -> str:
        return 'Betty'

    def _do_set_translatables(self) -> None:
        super()._do_set_translatables()
        self.new_project_action.setText(self._app.localizer._('New project...'))
        self.open_project_action.setText(self._app.localizer._('Open project...'))
        self._demo_action.setText(self._app.localizer._('View demo site...'))
        self.open_application_configuration_action.setText(self._app.localizer._('Settings...'))
        self.clear_caches_action.setText(self._app.localizer._('Clear all caches'))
        self.exit_action.setText(self._app.localizer._('Exit'))
        self.help_menu.setTitle('&' + self._app.localizer._('Help'))
        self.report_bug_action.setText(self._app.localizer._('Report a bug'))
        self.request_feature_action.setText(self._app.localizer._('Request a new feature'))
        self.about_action.setText(self._app.localizer._('About Betty'))

    @catch_exceptions
    def report_bug(self) -> None:
        body = f'''
## Summary

## Steps to reproduce

## Expected behavior

## System report
```
{report()}
```
'''.strip()
        webbrowser.open_new_tab('https://github.com/bartfeenstra/betty/issues/new?' + urlencode({
            'body': body,
            'labels': 'bug',
        }))

    @catch_exceptions
    def request_feature(self) -> None:
        body = '''
## Summary

## Expected behavior

'''.strip()
        webbrowser.open_new_tab('https://github.com/bartfeenstra/betty/issues/new?' + urlencode({
            'body': body,
            'labels': 'enhancement',
        }))

    @catch_exceptions
    def _about_betty(self) -> None:
        about_window = _AboutBettyWindow(self._app, self)
        about_window.show()

    @catch_exceptions
    def open_project(self) -> None:
        from betty.gui.project import ProjectWindow

        configuration_file_path_str, __ = QFileDialog.getOpenFileName(
            self,
            self._app.localizer._('Open your project from...'),
            '',
            get_configuration_file_filter().localize(self._app.localizer),
        )
        if not configuration_file_path_str:
            return
        wait(self._app.project.configuration.read(Path(configuration_file_path_str)))
        project_window = ProjectWindow(self._app)
        project_window.show()
        self.close()

    @catch_exceptions
    def new_project(self) -> None:
        from betty.gui.project import ProjectWindow

        configuration_file_path_str, __ = QFileDialog.getSaveFileName(
            self,
            self._app.localizer._('Save your new project to...'),
            '',
            get_configuration_file_filter().localize(self._app.localizer),
        )
        if not configuration_file_path_str:
            return
        configuration = ProjectConfiguration()
        wait(configuration.write(Path(configuration_file_path_str)))
        project_window = ProjectWindow(self._app)
        project_window.show()
        self.close()

    @catch_exceptions
    def _demo(self) -> None:
        serve_window = ServeDemoWindow(self._app, self)
        serve_window.show()

    @catch_exceptions
    @sync
    async def clear_caches(self) -> None:
        await self._app.cache.clear()

    @catch_exceptions
    def open_application_configuration(self) -> None:
        window = ApplicationConfiguration(self._app, self)
        window.show()


class _WelcomeText(Text):
    pass


class _WelcomeTitle(_WelcomeText):
    pass


class _WelcomeHeading(_WelcomeText):
    pass


class _WelcomeAction(QPushButton):
    pass


class WelcomeWindow(BettyMainWindow):
    # Allow the window to be as narrow as it can be.
    window_width = 1
    # This is a best guess at the minimum required height, because if we set this to 1, like the width, some of the
    # text will be clipped.
    window_height = 600

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)

        central_layout = QVBoxLayout()
        central_layout.addStretch()
        central_widget = QWidget()
        central_widget.setLayout(central_layout)
        self.setCentralWidget(central_widget)

        self._welcome = _WelcomeTitle()
        self._welcome.setAlignment(Qt.AlignmentFlag.AlignCenter)
        central_layout.addWidget(self._welcome)

        self._welcome_caption = _WelcomeText()
        central_layout.addWidget(self._welcome_caption)

        self._project_instruction = _WelcomeHeading()
        self._project_instruction.setAlignment(Qt.AlignmentFlag.AlignCenter)
        central_layout.addWidget(self._project_instruction)

        project_layout = QHBoxLayout()
        central_layout.addLayout(project_layout)

        self.open_project_button = _WelcomeAction(self)
        self.open_project_button.released.connect(self.open_project)
        project_layout.addWidget(self.open_project_button)

        self.new_project_button = _WelcomeAction(self)
        self.new_project_button.released.connect(self.new_project)
        project_layout.addWidget(self.new_project_button)

        self._demo_instruction = _WelcomeHeading()
        self._demo_instruction.setAlignment(Qt.AlignmentFlag.AlignCenter)
        central_layout.addWidget(self._demo_instruction)

        self.demo_button = _WelcomeAction(self)
        self.demo_button.released.connect(self._demo)
        central_layout.addWidget(self.demo_button)

    def _do_set_translatables(self) -> None:
        super()._do_set_translatables()
        self._welcome.setText(self._app.localizer._('Welcome to Betty'))
        self._welcome_caption.setText(self._app.localizer._('Betty helps you visualize and publish your family history by building interactive genealogy websites out of your <a href="{gramps_url}">Gramps</a> and <a href="{gedcom_url}">GEDCOM</a> family trees.').format(
            gramps_url='https://gramps-project.org/',
            gedcom_url='https://en.wikipedia.org/wiki/GEDCOM',
        ))
        self._project_instruction.setText(self._app.localizer._('Work on a new or existing site of your own'))
        self.open_project_button.setText(self._app.localizer._('Open an existing project'))
        self.new_project_button.setText(self._app.localizer._('Create a new project'))
        self._demo_instruction.setText(self._app.localizer._('View a demonstration of what a Betty site looks like'))
        self.demo_button.setText(self._app.localizer._('View a demo site'))


class _AboutBettyWindow(BettyWindow):
    window_width = 500
    window_height = 100

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)

        self._label = Text()
        self._label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.setCentralWidget(self._label)

    def _do_set_translatables(self) -> None:
        super()._do_set_translatables()
        self._label.setText(''.join(map(lambda x: '<p>%s</p>' % x, [
            self._app.localizer._('Version: {version}').format(
                version=about.version_label(),
            ),
            self._app.localizer._('Copyright 2019-{year} <a href="twitter.com/bartFeenstra">Bart Feenstra</a> & contributors. Betty is made available to you under the <a href="https://www.gnu.org/licenses/gpl-3.0.en.html">GNU General Public License, Version 3</a> (GPLv3).').format(
                year=datetime.now().year,
            ),
            self._app.localizer._('Follow Betty on <a href="https://twitter.com/Betty_Project">Twitter</a> and <a href="https://github.com/bartfeenstra/betty">Github</a>.'),
        ])))

    @property
    def title(self) -> str:
        return self._app.localizer._('About Betty')


class ApplicationConfiguration(BettyWindow):
    window_width = 400
    window_height = 150

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)

        self._form = QFormLayout()
        form_widget = QWidget()
        form_widget.setLayout(self._form)
        self.setCentralWidget(form_widget)
        locale_collector = TranslationsLocaleCollector(self._app, set(self._app.localizers.locales))
        for row in locale_collector.rows:
            self._form.addRow(*row)

    @property
    def title(self) -> str:
        return self._app.localizer._('Configuration')



from __future__ import annotations

from logging import getLogger
from os import path
from typing import Any, TypeVar

from PyQt6.QtCore import pyqtSlot, QObject
from PyQt6.QtGui import QIcon
from PyQt6.QtWidgets import QApplication, QWidget

from betty.app import App
from betty.error import UserFacingError
from betty.gui.error import ExceptionError, UnexpectedExceptionError
from betty.gui.locale import LocalizedWindow
from betty.locale import Str
from betty.serde.format import FormatRepository, FormatStr

QWidgetT = TypeVar('QWidgetT', bound=QWidget)


def get_configuration_file_filter() -> Str:
    formats = FormatRepository()
    return Str._(
        'Betty project configuration ({supported_formats})',
        supported_formats=FormatStr(formats.formats),
    )


class GuiBuilder:
    def gui_build(self) -> QWidget:
        raise NotImplementedError(repr(self))


def mark_valid(widget: QWidget) -> None:
    widget.setProperty('invalid', 'false')
    widget.setStyle(widget.style())
    widget.setToolTip('')


def mark_invalid(widget: QWidget, reason: str) -> None:
    widget.setProperty('invalid', 'true')
    widget.setStyle(widget.style())
    widget.setToolTip(reason)


class BettyWindow(LocalizedWindow):
    window_width = 800
    window_height = 600

    def __init__(self, app: App, *args: Any, **kwargs: Any):
        super().__init__(app, *args, **kwargs)
        self.resize(self.window_width, self.window_height)
        self.setWindowIcon(QIcon(path.join(path.dirname(__file__), 'assets', 'public', 'static', 'betty-512x512.png')))
        geometry = self.frameGeometry()
        screen = QApplication.primaryScreen()
        assert screen is not None
        geometry.moveCenter(screen.availableGeometry().center())
        self.move(geometry.topLeft())

    def _do_set_translatables(self) -> None:
        self.setWindowTitle(f'{self.title} - Betty')

    @property
    def title(self) -> str:
        raise NotImplementedError(repr(self))


class BettyApplication(QApplication):
    _STYLESHEET = """
        Caption {
            color: #333333;
            margin-bottom: 0.3em;
        }

        QLineEdit[invalid="true"] {
            border: 1px solid red;
            color: red;
        }

        QPushButton[pane-selector="true"] {
            padding: 10px;
        }

        LogRecord[level="50"],
        LogRecord[level="40"] {
            color: red;
        }

        LogRecord[level="30"] {
            color: yellow;
        }

        LogRecord[level="20"] {
            color: green;
        }

        LogRecord[level="10"],
        LogRecord[level="0"] {
            color: white;
        }

        _WelcomeText {
            padding: 10px;
        }

        _WelcomeTitle {
            font-size: 20px;
            padding: 10px;
        }

        _WelcomeHeading {
            font-size: 16px;
            margin-top: 50px;
        }

        _WelcomeAction {
            padding: 10px;
        }
        """

    def __init__(self, *args: Any, app: App, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.setApplicationName('Betty')
        self.setStyleSheet(self._STYLESHEET)
        self._app = app

    @pyqtSlot(
        Exception,
        QObject,
        bool,
    )
    def _catch_exception(
        self,
        e: Exception,
        parent: QObject,
        close_parent: bool,
    ) -> None:
        if isinstance(e, UserFacingError):
            window = ExceptionError(self._app, e, parent, close_parent=close_parent)
        else:
            getLogger().exception(e)
            window = UnexpectedExceptionError(self._app, e, parent, close_parent=close_parent)
        window.show()
